# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.ops.op import Operator
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import (
    Bool,
    Const,
    Date,
    Datetime,
    Int,
    String,
)


class StrUnary(Operator):
    def __init__(self, name: str, doc: str = ""):
        super().__init__(name, Signature(String(), return_type=String()), doc=doc)


str_strip = StrUnary(
    "str.strip",
    doc="""
Removes leading and trailing whitespace.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": ["  BCD ", "-- 00", " A^^u", "-O2"],
...         "b": ["12431", "transform", "12__*m", "   "],
...     },
...     name="string table",
... )
>>> t >> mutate(j=t.a.str.strip(), k=t.b.str.strip()) >> show()
Table string table, backend: PolarsImpl
shape: (4, 4)
┌────────┬───────────┬───────┬───────────┐
│ a      ┆ b         ┆ j     ┆ k         │
│ ---    ┆ ---       ┆ ---   ┆ ---       │
│ str    ┆ str       ┆ str   ┆ str       │
╞════════╪═══════════╪═══════╪═══════════╡
│   BCD  ┆ 12431     ┆ BCD   ┆ 12431     │
│ -- 00  ┆ transform ┆ -- 00 ┆ transform │
│  A^^u  ┆ 12__*m    ┆ A^^u  ┆ 12__*m    │
│ -O2    ┆           ┆ -O2   ┆           │
└────────┴───────────┴───────┴───────────┘
""",
)
str_upper = StrUnary(
    "str.upper",
    doc="""
Converts all alphabet letters to upper case.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": ["  BCD ", "-- 00", " A^^u", "-O2"],
...         "b": ["12431", "transform", "12__*m", "   "],
...     },
...     name="string table",
... )
>>> t >> mutate(j=t.a.str.upper(), k=t.b.str.upper()) >> show()
Table string table, backend: PolarsImpl
shape: (4, 4)
┌────────┬───────────┬────────┬───────────┐
│ a      ┆ b         ┆ j      ┆ k         │
│ ---    ┆ ---       ┆ ---    ┆ ---       │
│ str    ┆ str       ┆ str    ┆ str       │
╞════════╪═══════════╪════════╪═══════════╡
│   BCD  ┆ 12431     ┆   BCD  ┆ 12431     │
│ -- 00  ┆ transform ┆ -- 00  ┆ TRANSFORM │
│  A^^u  ┆ 12__*m    ┆  A^^U  ┆ 12__*M    │
│ -O2    ┆           ┆ -O2    ┆           │
└────────┴───────────┴────────┴───────────┘
""",
)
str_lower = StrUnary(
    "str.lower",
    doc="""
Converts all alphabet letters to lower case.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": ["  BCD ", "-- 00", " A^^u", "-O2"],
...         "b": ["12431", "transform", "12__*m", "   "],
...     },
...     name="string table",
... )
>>> t >> mutate(j=t.a.str.lower(), k=t.b.str.lower()) >> show()
Table string table, backend: PolarsImpl
shape: (4, 4)
┌────────┬───────────┬────────┬───────────┐
│ a      ┆ b         ┆ j      ┆ k         │
│ ---    ┆ ---       ┆ ---    ┆ ---       │
│ str    ┆ str       ┆ str    ┆ str       │
╞════════╪═══════════╪════════╪═══════════╡
│   BCD  ┆ 12431     ┆   bcd  ┆ 12431     │
│ -- 00  ┆ transform ┆ -- 00  ┆ transform │
│  A^^u  ┆ 12__*m    ┆  a^^u  ┆ 12__*m    │
│ -O2    ┆           ┆ -o2    ┆           │
└────────┴───────────┴────────┴───────────┘
""",
)

# We should write something about number of chars vs number of bytes here.
str_len = Operator(
    "str.len",
    Signature(String(), return_type=Int()),
    doc="""
Computes the length of the string.

Leading and trailing whitespace is included in the length.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": ["  BCD ", "-- 00", " A^^u", "-O2"],
...         "b": ["12431", "transform", "12__*m", "   "],
...     },
...     name="string table",
... )
>>> t >> mutate(j=t.a.str.len(), k=t.b.str.len()) >> show()
Table string table, backend: PolarsImpl
shape: (4, 4)
┌────────┬───────────┬─────┬─────┐
│ a      ┆ b         ┆ j   ┆ k   │
│ ---    ┆ ---       ┆ --- ┆ --- │
│ str    ┆ str       ┆ i64 ┆ i64 │
╞════════╪═══════════╪═════╪═════╡
│   BCD  ┆ 12431     ┆ 6   ┆ 5   │
│ -- 00  ┆ transform ┆ 5   ┆ 9   │
│  A^^u  ┆ 12__*m    ┆ 5   ┆ 6   │
│ -O2    ┆           ┆ 3   ┆ 3   │
└────────┴───────────┴─────┴─────┘
""",
)

str_replace_all = Operator(
    "str.replace_all",
    Signature(String(), Const(String()), Const(String()), return_type=String()),
    param_names=["self", "substr", "replacement"],
    doc="""
Replaces all occurrences of a given substring by a different string.

:param substr:
    The string to replace.

:param replacement:
    The replacement string.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": ["  BCD ", "-- 00", " A^^u", "-O2", ""],
...         "b": ["12431", "transform", "12__*m", "   ", "abbabbabba"],
...     },
...     name="string table",
... )
>>> (
...     t
...     >> mutate(
...         r=t.a.str.replace_all("-", "?"),
...         s=t.b.str.replace_all("ansf", "[---]"),
...         u=t.b.str.replace_all("abba", "#"),
...     )
...     >> show()
... )
Table string table, backend: PolarsImpl
shape: (5, 5)
┌────────┬────────────┬────────┬────────────┬───────────┐
│ a      ┆ b          ┆ r      ┆ s          ┆ u         │
│ ---    ┆ ---        ┆ ---    ┆ ---        ┆ ---       │
│ str    ┆ str        ┆ str    ┆ str        ┆ str       │
╞════════╪════════════╪════════╪════════════╪═══════════╡
│   BCD  ┆ 12431      ┆   BCD  ┆ 12431      ┆ 12431     │
│ -- 00  ┆ transform  ┆ ?? 00  ┆ tr[---]orm ┆ transform │
│  A^^u  ┆ 12__*m     ┆  A^^u  ┆ 12__*m     ┆ 12__*m    │
│ -O2    ┆            ┆ ?O2    ┆            ┆           │
│        ┆ abbabbabba ┆        ┆ abbabbabba ┆ #bb#      │
└────────┴────────────┴────────┴────────────┴───────────┘
""",
)

str_starts_with = Operator(
    "str.starts_with",
    Signature(String(), Const(String()), return_type=Bool()),
    param_names=["self", "prefix"],
    doc="""
Whether the string starts with a given prefix.

:param prefix:
    The prefix to check.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": ["  BCD ", "-- 00", " A^^u", "-O2", ""],
...         "b": ["12431", "transform", "12__*m", "   ", "abbabbabba"],
...     },
...     name="string table",
... )
>>> (
...     t
...     >> mutate(
...         j=t.a.str.starts_with("-"),
...         k=t.b.str.starts_with("12"),
...     )
...     >> show()
... )
Table string table, backend: PolarsImpl
shape: (5, 4)
┌────────┬────────────┬───────┬───────┐
│ a      ┆ b          ┆ j     ┆ k     │
│ ---    ┆ ---        ┆ ---   ┆ ---   │
│ str    ┆ str        ┆ bool  ┆ bool  │
╞════════╪════════════╪═══════╪═══════╡
│   BCD  ┆ 12431      ┆ false ┆ true  │
│ -- 00  ┆ transform  ┆ true  ┆ false │
│  A^^u  ┆ 12__*m     ┆ false ┆ true  │
│ -O2    ┆            ┆ true  ┆ false │
│        ┆ abbabbabba ┆ false ┆ false │
└────────┴────────────┴───────┴───────┘
""",
)

str_ends_with = Operator(
    "str.ends_with",
    Signature(String(), Const(String()), return_type=Bool()),
    param_names=["self", "suffix"],
    doc="""
Whether the string ends with a given suffix.

:param suffix:
    The suffix to check.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": ["  BCD ", "-- 00", " A^^u", "-O2", ""],
...         "b": ["12431", "transform", "12__*m", "   ", "abbabbabba"],
...     },
...     name="string table",
... )
>>> (
...     t
...     >> mutate(
...         j=t.a.str.ends_with(""),
...         k=t.b.str.ends_with("m"),
...         l=t.a.str.ends_with("^u"),
...     )
...     >> show()
... )
Table string table, backend: PolarsImpl
shape: (5, 5)
┌────────┬────────────┬──────┬───────┬───────┐
│ a      ┆ b          ┆ j    ┆ k     ┆ l     │
│ ---    ┆ ---        ┆ ---  ┆ ---   ┆ ---   │
│ str    ┆ str        ┆ bool ┆ bool  ┆ bool  │
╞════════╪════════════╪══════╪═══════╪═══════╡
│   BCD  ┆ 12431      ┆ true ┆ false ┆ false │
│ -- 00  ┆ transform  ┆ true ┆ true  ┆ false │
│  A^^u  ┆ 12__*m     ┆ true ┆ true  ┆ true  │
│ -O2    ┆            ┆ true ┆ false ┆ false │
│        ┆ abbabbabba ┆ true ┆ false ┆ false │
└────────┴────────────┴──────┴───────┴───────┘
""",
)


str_contains = Operator(
    "str.contains",
    Signature(String(), Const(String()), Const(Bool()), Const(Bool()), return_type=Bool()),
    param_names=["self", "pattern", "allow_regex", "true_if_regex_unsupported"],
    default_values=[..., ..., True, False],
    doc="""
Whether the string contains a given pattern or substring.

:param pattern:
    The pattern or substring to look for.

:param allow_regex:
    If set to `True` (the default value), `pattern` is treated as a regular
    expression. Otherwise, `pattern` is treated as a literal string.

:param true_if_regex_unsupported:
    If set to `True`, this function always returns `True` on backends that do
    not support regex.

Note
----
If the backend does not support regex, `pattern` is automatically treated as a
literal string, regardless of `allow_regex`.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": ["  BCD ", "-- 00", " A^^u", "-O2", ""],
...         "b": ["12431", "transform", "12__*m", "   ", "abbabbabba"],
...     },
...     name="string table",
... )
>>> (
...     t
...     >> mutate(
...         j=t.a.str.contains(" "),
...         k=t.b.str.contains("a"),
...         l=t.b.str.contains(""),
...     )
...     >> show()
... )
Table string table, backend: PolarsImpl
shape: (5, 5)
┌────────┬────────────┬───────┬───────┬──────┐
│ a      ┆ b          ┆ j     ┆ k     ┆ l    │
│ ---    ┆ ---        ┆ ---   ┆ ---   ┆ ---  │
│ str    ┆ str        ┆ bool  ┆ bool  ┆ bool │
╞════════╪════════════╪═══════╪═══════╪══════╡
│   BCD  ┆ 12431      ┆ true  ┆ false ┆ true │
│ -- 00  ┆ transform  ┆ true  ┆ true  ┆ true │
│  A^^u  ┆ 12__*m     ┆ true  ┆ false ┆ true │
│ -O2    ┆            ┆ false ┆ false ┆ true │
│        ┆ abbabbabba ┆ false ┆ true  ┆ true │
└────────┴────────────┴───────┴───────┴──────┘
""",
)

str_slice = Operator(
    "str.slice",
    Signature(String(), Int(), Int(), return_type=String()),
    param_names=["self", "offset", "n"],
    doc="""
Returns a substring of the input string.

:param offset:
    The 0-based index of the first character included in the result.

:param n:
    The number of characters to include. If the string is shorter than *offset*
    + *n*, the result only includes as many characters as there are.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": ["  BCD ", "-- 00", " A^^u", "-O2", ""],
...         "b": ["12431", "transform", "12__*m", "   ", "abbabbabba"],
...     },
...     name="string table",
... )
>>> (
...     t
...     >> mutate(
...         j=t.a.str.slice(0, 2),
...         k=t.b.str.slice(4, 10),
...     )
...     >> show()
... )
Table string table, backend: PolarsImpl
shape: (5, 4)
┌────────┬────────────┬─────┬────────┐
│ a      ┆ b          ┆ j   ┆ k      │
│ ---    ┆ ---        ┆ --- ┆ ---    │
│ str    ┆ str        ┆ str ┆ str    │
╞════════╪════════════╪═════╪════════╡
│   BCD  ┆ 12431      ┆     ┆ 1      │
│ -- 00  ┆ transform  ┆ --  ┆ sform  │
│  A^^u  ┆ 12__*m     ┆  A  ┆ *m     │
│ -O2    ┆            ┆ -O  ┆        │
│        ┆ abbabbabba ┆     ┆ bbabba │
└────────┴────────────┴─────┴────────┘
""",
)

str_to_datetime = Operator("str.to_datetime", Signature(String(), return_type=Datetime()))

str_to_date = Operator("str.to_date", Signature(String(), return_type=Date()))



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform.extended import *
from tests.util import assert_result_equal


def test_noop(df2):
    assert_result_equal(df2, lambda t: t >> filter())
    assert_result_equal(df2, lambda t: t >> filter(t.col1 == t.col1))


def test_simple_filter(df2):
    assert_result_equal(df2, lambda t: t >> filter(t.col1 == 2))
    assert_result_equal(df2, lambda t: t >> filter(t.col1 != 2))


def test_chained_filters(df2):
    assert_result_equal(df2, lambda t: t >> filter(1 < t.col1) >> filter(t.col1 < 5))

    assert_result_equal(df2, lambda t: t >> filter(1 < t.col1) >> filter(t.col3 < 0.25))


def test_filter_empty_result(df3):
    assert_result_equal(
        df3,
        lambda t: t >> filter(t.col1 == 0) >> filter(t.col2 == 2) >> filter(t.col4 < 2),
    )


def test_filter_after_mutate(df4):
    assert_result_equal(
        df4,
        lambda t: t >> mutate(x=t.col1 <= t.col2) >> filter(C.x),
    )


def test_filter_is_in(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> filter(
            C.col1.is_in(0, 2),
            C.col2.is_in(0, t.col1 * t.col2),
        )
        >> mutate(u=C.col3.is_in()),
    )

    assert_result_equal(
        df4,
        lambda t: t >> filter((-(t.col4 // 2 - 1)).is_in(1, 4, t.col1 + t.col2)),
    )

    assert_result_equal(df4, lambda t: t >> filter(t.col1.is_in(None)))

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(x=t.col1.is_in(0, 2))
        >> filter(
            t.col2.is_in(0, 2) & C.x,
        ),
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy
import dataclasses
from pprint import pformat
from typing import Optional
from uuid import UUID

from pydiverse.transform._internal import errors
from pydiverse.transform._internal.backend.table_impl import TableImpl
from pydiverse.transform._internal.ops.op import Ftype
from pydiverse.transform._internal.tree import types, verbs
from pydiverse.transform._internal.tree.ast import AstNode
from pydiverse.transform._internal.tree.col_expr import Col, ColFn


@dataclasses.dataclass(slots=True)
class Cache:
    name_to_uuid: dict[str, UUID]  # the selected columns, in order
    uuid_to_name: dict[UUID, str]  # again, only the selected columns + in order
    partition_by: list[UUID]
    derived_from: set[AstNode]  # for detecting invalid columns and self-joins
    cols: dict[UUID, Col]  # all columns in current scope (including hidden ones)

    # the following are only necessary for subquery detection
    limit: int
    group_by: set[UUID]
    is_filtered: bool

    backend: type[TableImpl]

    def __repr__(self) -> str:
        return (
            "Cache(\n"
            + "  name_to_uuid="
            + f"""{
                "{" + pformat(self.name_to_uuid, indent=16)[16:]
                if len(self.name_to_uuid) > 1
                else pformat(self.name_to_uuid)
            }"""
            + ",\n"
            + "  uuid_to_name="
            + f"""{
                "{" + pformat(self.uuid_to_name, indent=16)[16:]
                if len(self.uuid_to_name) > 1
                else pformat(self.name_to_uuid)
            }"""
            + ",\n"
            + f"  partition_by={self.partition_by}\n"
            + "  derived_from={"
            + f"{', '.join(d.short_name() for d in self.derived_from)}"
            + "},\n"
            + "  cols={"
            + f"""{
                pformat(
                    {uid: col.ast_repr(depth=0) for uid, col in self.cols.items()},
                    indent=8,
                )[8:]
            }"""
            + ",\n"
            + f"  limit={self.limit},\n"
            + f"  group_by={self.group_by},\n"
            + f"  is_filtered={self.is_filtered},\n)"
        )

    # For a column to be usable in an expression, the table it comes from must be an
    # ancestor of the current table AND the column's UUID must be in `all_cols` of the
    # current table (the latter is necessary because not every column is usable after
    # `summarize``)

    @staticmethod
    def from_ast(node: AstNode) -> "Cache":
        if isinstance(node, verbs.Verb):
            if isinstance(node, verbs.Join | verbs.Union):
                return Cache.from_ast(node.child).update(node, right_cache=Cache.from_ast(node.right))
            else:
                return Cache.from_ast(node.child).update(node)

        assert isinstance(node, TableImpl)
        return Cache(
            name_to_uuid={col.name: col._uuid for col in node.cols.values()},
            uuid_to_name={col._uuid: col.name for col in node.cols.values()},
            partition_by=[],
            derived_from={node},
            cols={col._uuid: col for col in node.cols.values()},
            limit=0,
            group_by=set(),
            is_filtered=False,
            backend=type(node),
        )

    def update(self, node: verbs.Verb, *, right_cache: Optional["Cache"] = None) -> "Cache":
        """
        Returns a new cache for `node`, assuming `self` is the cache of `node.child`.
        Does not modify `self`.
        """

        res = copy.copy(self)

        if isinstance(node, verbs.Alias):
            if node.uuid_map is not None:
                res.name_to_uuid = {name: node.uuid_map[uid] for name, uid in self.name_to_uuid.items()}
                res.uuid_to_name = {uid: name for name, uid in res.name_to_uuid.items()}
                res.cols = {
                    node.uuid_map[uid]: Col(col.name, node, node.uuid_map[uid], col._dtype, col._ftype)
                    for uid, col in self.cols.items()
                }
                res.partition_by = [node.uuid_map[uid] for uid in self.partition_by]
                res.derived_from = set()

        elif isinstance(node, verbs.Select):
            selected_uuids = set(col._uuid for col in node.select)
            res.uuid_to_name = {uid: name for uid, name in self.uuid_to_name.items() if uid in selected_uuids}
            res.name_to_uuid = {name: uid for uid, name in res.uuid_to_name.items()}

        elif isinstance(node, verbs.Rename):
            res.name_to_uuid = {
                (new_name if (new_name := node.name_map.get(name)) else name): uid
                for name, uid in self.name_to_uuid.items()
            }
            res.uuid_to_name = {uid: name for name, uid in res.name_to_uuid.items()}

        elif isinstance(node, verbs.Mutate):
            res.cols = self.cols | {
                uid: Col(name, node, uid, val.dtype(), val.ftype(agg_is_window=True))
                for name, val, uid in zip(node.names, node.values, node.uuids, strict=True)
            }
            res.name_to_uuid = self.name_to_uuid | {name: uid for name, uid in zip(node.names, node.uuids, strict=True)}
            res.uuid_to_name = {uid: name for name, uid in res.name_to_uuid.items()}

        elif isinstance(node, verbs.Filter):
            res.is_filtered = True

        elif isinstance(node, verbs.GroupBy):
            res.partition_by = [col._uuid for col in node.group_by]
            if node.add:
                res.partition_by = self.partition_by + res.partition_by

        elif isinstance(node, verbs.Ungroup):
            res.partition_by = []

        elif isinstance(node, verbs.Summarize):
            overwritten = {col_name for col_name in node.names if col_name in self.name_to_uuid}
            cols = {
                self.uuid_to_name[uid]: self.cols[uid]
                for uid in self.partition_by
                if self.uuid_to_name[uid] not in overwritten
            } | {
                name: Col(name, node, uid, val.dtype(), val.ftype())
                for name, val, uid in zip(node.names, node.values, node.uuids, strict=True)
            }

            res.cols = {col._uuid: col for _, col in cols.items()}
            res.name_to_uuid = {name: col._uuid for name, col in cols.items()}
            res.uuid_to_name = {uid: name for name, uid in res.name_to_uuid.items()}
            res.group_by = res.group_by | set(res.partition_by)
            res.partition_by = []

        elif isinstance(node, verbs.SliceHead):
            res.limit = node.n

        elif isinstance(node, verbs.Join):
            assert right_cache is not None

            res.cols = self.cols | right_cache.cols
            res.name_to_uuid = self.name_to_uuid | right_cache.name_to_uuid
            res.uuid_to_name = {uid: name for name, uid in res.name_to_uuid.items()}

            res.derived_from = self.derived_from | right_cache.derived_from
            res.limit = 0
            res.group_by = set()

        elif isinstance(node, verbs.Union):
            assert right_cache is not None

            # For union, visible columns must match (validated in verb function)
            # Hidden columns: are removed (we don't keep names for them and it is unlike they match in uuid)
            res.cols = {uid: col for uid, col in self.cols.items() if uid in self.uuid_to_name}
            # Visible columns should match, so we keep left table's name_to_uuid
            # (right table's visible columns are the same by validation)
            res.name_to_uuid = self.name_to_uuid.copy()
            res.uuid_to_name = self.uuid_to_name.copy()

            res.derived_from = self.derived_from | right_cache.derived_from
            res.limit = 0
            res.group_by = set()

        elif isinstance(node, verbs.SubqueryMarker):
            res.cols = {
                uid: Col(
                    col.name,
                    node.child,
                    uid,
                    types.without_const(col._dtype),
                    Ftype.ELEMENT_WISE,
                )
                for uid, col in self.cols.items()
            }
            res.limit = 0
            res.group_by = set()
            res.is_filtered = False

        assert len(res.name_to_uuid) == len(res.uuid_to_name)
        res.derived_from = res.derived_from | {node}

        return res

    # Returns whether applying `node` to this table would require it to be wrapped in a
    # subquery. (so `self` is the old cache and `node` the new AST).
    # Either returns a reason for the subquery or None.
    def requires_subquery(self, node: verbs.Verb) -> str | None:
        if self.backend.backend_name == "polars":
            return None

        if (
            isinstance(
                node,
                verbs.Filter | verbs.Summarize | verbs.Arrange | verbs.GroupBy | verbs.Join | verbs.Union,
            )
            and self.limit != 0
        ):
            return f"`{node.__class__.__name__.lower()}` after `slice_head`"

        if isinstance(node, verbs.Mutate) and any(
            any(
                col.ftype(agg_is_window=True) in (Ftype.WINDOW, Ftype.AGGREGATE)
                for col in fn.iter_subtree_postorder()
                if isinstance(col, Col)
            )
            for fn in node.iter_col_nodes()
            if (isinstance(fn, ColFn) and fn.op.ftype in (Ftype.AGGREGATE, Ftype.WINDOW))
        ):
            return "nested window / aggregation functions in `mutate`"

        if isinstance(node, verbs.Filter) and any(
            col.ftype(agg_is_window=True) == Ftype.WINDOW for col in node.iter_col_nodes() if isinstance(col, Col)
        ):
            return "window function in `filter`"

        if isinstance(node, verbs.Summarize):
            if self.group_by and self.group_by != set(self.partition_by):
                return "nested summarize"
            if any(
                (col.ftype(agg_is_window=False) in (Ftype.WINDOW, Ftype.AGGREGATE))
                for col in node.iter_col_nodes()
                if isinstance(col, Col)
            ):
                return "nested window / aggregation functions in `summarize`"
            if any(self.cols[uid].ftype() == Ftype.WINDOW for uid in self.partition_by):
                return "window function among grouping columns"

        if isinstance(node, verbs.Join):
            if self.group_by:
                return "join with a grouped table"

            if (node.how == "full" or (node.child not in self.derived_from and node.how == "left")) and any(
                types.is_const(self.cols[uid].dtype()) for uid in self.uuid_to_name.keys()
            ):
                return "left / full join with a table containing a constant column"

            if any(self.cols[uid].ftype() == Ftype.WINDOW for uid in self.uuid_to_name.keys()):
                return "join with a table containing window function expression"

            if any(
                col.ftype() != Ftype.ELEMENT_WISE and col._uuid in self.cols
                for col in node.on.iter_subtree_postorder()
                if isinstance(col, Col)
            ):
                return "window / aggregation functions in join condition"

            if self.is_filtered and node.how == "full":
                return "full join with a filtered table"

        if isinstance(node, verbs.Union):
            if self.group_by:
                return "union with a grouped table"

            if any(self.cols[uid].ftype() == Ftype.WINDOW for uid in self.uuid_to_name.keys()):
                return "union with a table containing window function expression"

        return None

    def selected_cols(self) -> list[Col]:
        return [self.cols[uid] for uid in self.uuid_to_name.keys()]


def transfer_col_references(table, ref_source):
    """
    Transfers the column references from `ref_source` to `table`.

    The returned table contains all selected columns of `table`, but its columns are
    now referenced by the columns from `ref_source`. All column names selected in
    `table` must also be present in `ref_source`.

    :param table:
        The table from which the data is taken.

    :param ref_source:
        The table from which the column references are taken.


    Examples
    --------
    **Materialization without breaking the functional flow.** Say you have a function
    `your_materialization_fn` that writes a transform table to a database and returns a
    transform table again. Then you can define a custom verb

    >>> @verb
    ... def materialize(table) -> pdt.Table:
    ...     new = your_materialization_fn(table)
    ...     return pdt.transfer_col_references(new, table)

    With this verb, it is possible to write

    >>> t = pdt.Table(dict(a=[1, 2, 5], b=["x", "y", "z"]), name="t")
    >>> t >> filter(t.a >= 2) >> materialize() >> mutate(z=t.a + t.b.str.len())
    Table `t` (backend: polars)
    shape: (2, 3)
    ┌─────┬─────┬─────┐
    │ a   ┆ b   ┆ z   │
    │ --- ┆ --- ┆ --- │
    │ i64 ┆ str ┆ i64 │
    ╞═════╪═════╪═════╡
    │ 2   ┆ y   ┆ 3   │
    │ 5   ┆ z   ┆ 6   │
    └─────┴─────┴─────┘

    Without `transfer_col_references`, it would not be possible to use `t.a` and `t.b`
    in the `mutate`. (Of course, you would normally have a SQL backend when
    materializing, not a polars backend like in the example here.)
    """
    from pydiverse.transform._internal.pipe.table import Table
    from pydiverse.transform._internal.tree.verbs import Alias

    errors.check_arg_type(Table, "transfer_col_references", "table", table)
    errors.check_arg_type(Table, "transfer_col_references", "ref_source", ref_source)

    if (col := next((col for col in table if col.name not in ref_source), None)) is not None:
        raise ValueError(
            f"column {col.ast_repr()} of the table `{table._ast.short_name()}` does "
            "not exist in the reference source table "
            f"`{ref_source._ast.short_name()}`"
        )

    new = copy.copy(table)
    new._ast = Alias(
        new._ast,
        uuid_map={uid: ref_source._cache.name_to_uuid[name] for uid, name in table._cache.uuid_to_name.items()},
    )
    new._cache = table._cache.update(new._ast)

    return new



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

# ruff: noqa: F405

import random
import string
from functools import partial

import numpy as np
import polars as pl
from polars.testing import assert_frame_equal

import pydiverse.transform as pdt
from pydiverse.transform._internal.ops import ops
from pydiverse.transform._internal.ops.op import Ftype, Operator
from pydiverse.transform._internal.ops.ops.markers import Marker
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.col_expr import ColFn
from pydiverse.transform._internal.tree.types import Tyvar
from pydiverse.transform.common import *  # noqa: F403
from tests.util.backend import BACKEND_TABLES

rng = np.random.default_rng()
letters = list(string.printable)

ALL_TYPES = [pdt.Int(), pdt.Float(), pdt.Bool(), pdt.String()]
MEAN_HEIGHT = 3

RNG_FNS = {
    pdt.Float(): rng.standard_normal,
    pdt.Int(): partial(rng.integers, -(1 << 13), 1 << 13),
    pdt.Bool(): partial(rng.integers, 0, 1, dtype=bool),
    pdt.String(): (lambda rows: np.array(["".join(random.choices(letters, k=rng.poisson(10))) for _ in range(rows)])),
}


def gen_table(rows: int, types: dict[pdt.Dtype, int]) -> pl.DataFrame:
    d = pl.DataFrame()

    for ty, fn in RNG_FNS.items():
        if ty in types:
            d = d.with_columns(
                **{f"{ty.__class__.__name__.lower()} #{i + 1}": pl.lit(fn(rows)) for i in range(types[ty])}
            )

    return d


ops_with_return_type: dict[pdt.Dtype, list[tuple[Operator, Signature]]] = {ty: [] for ty in ALL_TYPES}

for op in ops.__dict__.values():
    if not isinstance(op, Operator) or op.ftype != Ftype.ELEMENT_WISE or isinstance(op, Marker):
        continue
    for sig in op.signatures:
        if not all(t in (*ALL_TYPES, Tyvar("T")) for t in (*sig.types, sig.return_type)):
            continue

        if isinstance(sig.return_type, Tyvar) or any(isinstance(param, Tyvar) for param in sig.types):
            for ty in ALL_TYPES:
                rtype = ty if isinstance(sig.return_type, Tyvar) else sig.return_type
                ops_with_return_type[rtype].append(
                    (
                        op,
                        Signature(
                            *(ty if isinstance(param, Tyvar) else param for param in sig.types),
                            return_type=rtype,
                        ),
                    )
                )
        else:
            ops_with_return_type[sig.return_type].append((op, sig))


def gen_expr(dtype: pdt.Dtype, cols: dict[pdt.Dtype, list[str]], q: float = 0.0) -> pdt.ColExpr:
    if dtype.const:
        return RNG_FNS[dtype.without_const()](1).item()

    if q > 1:
        # we always use C here so the expression does not have to be generated for each
        # backend
        return C[rng.choice(cols[dtype])]

    op, sig = rng.choice(ops_with_return_type[dtype])
    assert isinstance(op, Operator)
    assert isinstance(sig, Signature)

    args = []
    for param in sig.types[: len(sig.types) - sig.is_vararg]:
        args.append(gen_expr(param, cols, q + rng.exponential(1 / MEAN_HEIGHT)))

    if sig.is_vararg:
        nargs = int(rng.normal(2.5, 1 / 1.5))
        for _ in range(nargs):
            args.append(gen_expr(sig.types[-1], cols, q + rng.exponential(1 / MEAN_HEIGHT)))

    return ColFn(op, *args)


it = int(input("number of iterations: "))
rows = int(input("number of rows: "))
seed = int(input("seed: "))

rng = np.random.default_rng(seed)
NUM_COLS_PER_TYPE = 5

df = gen_table(rows, {dtype: NUM_COLS_PER_TYPE for dtype in ALL_TYPES})


tables = {backend: fn(df, "t") for backend, fn in BACKEND_TABLES.items()}
cols = {dtype: [col.name for col in tables["polars"] if col.dtype() <= dtype] for dtype in ALL_TYPES}

for _ in range(it):
    expr = gen_expr(rng.choice(ALL_TYPES), cols)
    results = {backend: table >> mutate(y=expr) >> select(C.y) >> export(Polars()) for backend, table in tables.items()}
    for _backend, res in results:
        assert_frame_equal(results["polars"], res)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.transform as pdt
from pydiverse.transform.extended import *
from tests.fixtures.backend import skip_backends
from tests.util import assert_result_equal


def test_union_basic(df3, df4):
    """Test basic union with only visible columns."""
    # Test pipe syntax: tbl1 >> union(tbl2)
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> union(u),
        check_row_order=False,
    )

    # Test direct call syntax: union(tbl1, tbl2)
    assert_result_equal(
        (df3, df4),
        lambda t, u: union(t, u),
        check_row_order=False,
    )


def test_union_all(df3, df4):
    """Test UNION ALL (keeping duplicates)."""
    # Test pipe syntax with distinct=False (keeps duplicates)
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> union(u, distinct=False),
        check_row_order=False,
    )

    # Test direct call syntax with distinct=False (keeps duplicates)
    assert_result_equal(
        (df3, df4),
        lambda t, u: union(t, u, distinct=False),
        check_row_order=False,
    )


@skip_backends("ibm_db2")
def test_union_distinct(df3, df4):
    """Test UNION (removing duplicates)."""
    # Test pipe syntax with distinct=True (default, removes duplicates)
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> union(u, distinct=True),
        check_row_order=False,
    )

    # Test direct call syntax with distinct=True (default, removes duplicates)
    assert_result_equal(
        (df3, df4),
        lambda t, u: union(t, u, distinct=True),
        check_row_order=False,
    )


def test_union_distinct2(df3, df4):
    """Test UNION (removing duplicates)."""
    # Test pipe syntax with distinct=True (default, removes duplicates)
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> drop(t.col7) >> union(u >> drop(u.col7), distinct=True),
        check_row_order=False,
    )

    # Test direct call syntax with distinct=True (default, removes duplicates)
    assert_result_equal(
        (df3, df4),
        lambda t, u: union(t >> drop(t.col7), u >> drop(u.col7), distinct=True),
        check_row_order=False,
    )


def test_union_with_select(df3, df4):
    """Test union with selected columns."""
    # Test with select on left table
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> select(t.col1) >> union(u >> select(u.col1)),
        check_row_order=False,
    )

    # Test with select on right table
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> select(t.col2) >> union(u >> select(u.col2)),
        check_row_order=False,
    )


def test_union_with_hidden_columns_left(df3, df4):
    """Test union when left table has hidden columns."""
    # Left table has hidden column, right table has all columns visible
    assert_result_equal(
        (df3, df4),
        lambda t, u: t
        >> mutate(hidden_col=t.col1 * 10)  # create a column
        >> select(t.col1, t.col2)  # hidden_col becomes hidden
        >> union(u >> select(u.col1, u.col2)),  # right has no hidden columns
        check_row_order=False,
    )


def test_union_with_hidden_columns_right(df3, df4):
    """Test union when right table has hidden columns that left table doesn't have."""
    # Right table has hidden column, left table has all columns visible
    # The hidden column in right should not appear in the union result
    assert_result_equal(
        (df3, df4),
        lambda t, u: t
        >> select(t.col1, t.col2)  # all visible
        >> union(u >> mutate(hidden_col=u.col1 * 10) >> select(u.col1, u.col2)),  # hidden_col becomes hidden in right
        check_row_order=False,
    )


def test_union_with_hidden_columns_partial_match(df3, df4):
    """Test union when both tables have hidden columns that partially match."""
    # Both tables have some hidden columns in common, some different
    # Only hidden columns that exist in BOTH tables should be preserved
    assert_result_equal(
        (df3, df4),
        lambda t, u: t
        >> mutate(shared_hidden=t.col1 * 2, left_only=t.col1 + 100)  # create columns
        >> select(t.col1, t.col2)  # shared_hidden and left_only become hidden
        >> union(
            u
            >> mutate(shared_hidden=u.col1 * 2, right_only=u.col1 + 200)
            >> select(u.col1, u.col2)  # shared_hidden and right_only become hidden
        ),
        check_row_order=False,
    )


def test_union_with_mutate_hidden(df3, df4):
    """Test union with mutated columns that become hidden."""
    # Create columns in mutate, then hide some
    assert_result_equal(
        (df3, df4),
        lambda t, u: t
        >> mutate(x=t.col1 * 2, y=t.col2 * 10)
        >> select(t.col1, C.x)  # y becomes hidden
        >> union(u >> mutate(x=u.col1 * 2, y=u.col2 * 10) >> select(u.col1, C.x)),
        check_row_order=False,
    )


def test_union_chained(df3, df4):
    """Test chaining multiple unions."""
    # Chain multiple unions together
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> union(u) >> union(t),
        check_row_order=False,
    )


@skip_backends("sqlite")  # sqlite only supports UNION for trivial queries
def test_union_after_operations(df3, df4):
    """Test union after other operations like filter and arrange."""
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> filter(t.col1 > 0) >> arrange(t.col1) >> union(u >> filter(u.col1 > 0) >> arrange(u.col1)),
        check_row_order=False,
    )


def test_union_error_different_columns(df3, df4):
    """Test that union raises error when columns don't match."""
    # Should raise ValueError when column names don't match
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> select(t.col1) >> union(u >> select(u.col2)),
        exception=ValueError,
    )


def test_union_error_different_backends():
    """Test that union raises error when backends don't match."""
    import polars as pl
    import sqlalchemy as sqa

    # Create tables with different backends
    polars_tbl = pdt.Table(pl.DataFrame({"a": [1, 2]}))
    engine = sqa.create_engine("sqlite:///:memory:")
    pl.DataFrame({"a": [1, 2]}).write_database("test", engine, if_table_exists="replace")
    sql_tbl = pdt.Table("test", pdt.SqlAlchemy(engine))

    # Should raise TypeError
    with pytest.raises(TypeError, match="cannot union two tables with different backends"):
        polars_tbl >> union(sql_tbl)


def test_union_error_grouped_table(df3, df4):
    """Test that union raises error when table is grouped."""
    # Should raise ValueError when trying to union a grouped table
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> group_by(t.col1) >> union(u),
        exception=ValueError,
    )

    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> union(u >> group_by(u.col1)),
        exception=ValueError,
    )


def test_union_with_rename(df3, df4):
    """Test union after renaming columns."""
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> rename({"col1": "a", "col2": "b"}) >> union(u >> rename({"col1": "a", "col2": "b"})),
        check_row_order=False,
    )


def test_union_empty_tables():
    """Test union with empty tables."""
    import polars as pl

    empty1 = pdt.Table(pl.DataFrame({"a": [], "b": []}))
    empty2 = pdt.Table(pl.DataFrame({"a": [], "b": []}))

    result = empty1 >> union(empty2) >> export(pdt.Polars(lazy=False))
    assert len(result) == 0
    assert result.columns == ["a", "b"]


def test_union_different_column_order1(df3, df4):
    """Test union when columns are in different order (should reorder automatically)."""
    # Columns in different order should still work - union should handle reordering
    # Note: This tests that the backend correctly reorders columns to match
    assert_result_equal(
        (df3, df4),
        lambda t, u: t >> select(t.col1, t.col2) >> union(u >> select(u.col1, u.col2)),  # same columns, same order
        check_row_order=False,
    )


def test_union_different_column_order2(df3, df4):
    # Test that union handles column reordering correctly
    # The backend should reorder right table columns to match left order (by column name)
    assert_result_equal(
        (df3, df4),
        lambda t, u: t
        >> select(t.col1, t.col2)
        >> union(u >> select(u.col2, u.col1)),  # different order - should be reordered
        check_row_order=False,
    )


def test_union_different_column_order3(df3, df4):
    # Test that union handles column reordering correctly
    # The backend should not reorder right table columns since order matches based on name
    assert_result_equal(
        (df3, df4),
        lambda t, u: t
        >> select(t.col1, t.col2)
        >> union(u >> select(u.col2, u.col1) >> rename({u.col1: "col2", u.col2: "col1"})),
        check_row_order=False,
    )


def test_union_different_column_order4(df3, df4):
    # Test that union handles column reordering correctly
    # The backend should reorder right table columns to match left order
    assert_result_equal(
        (df3, df4),
        lambda t, u: t
        >> select(t.col1, t.col2)
        >> union(u >> select(u.col1, u.col2) >> rename({u.col1: "col2", u.col2: "col1"})),
        check_row_order=False,
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pydiverse.transform as pdt
from pydiverse.transform import C, mutate, summarize
from tests.util.assertion import assert_result_equal

_ = pdt


def test_add(df_int):
    assert_result_equal(
        df_int,
        lambda t: t >> (lambda s: mutate(**{f"add_{c.name}_{d.name}": c + d for d in s for c in s})),
    )


def test_sub(df_int):
    assert_result_equal(
        df_int,
        lambda t: t
        >> (lambda s: (mutate(**{f"sub_{c.name}_{d.name}": c - d for d in s for c in s}) >> (lambda u: mutate()))),
    )


def test_neg(df_int):
    assert_result_equal(
        df_int,
        lambda t: t >> (lambda s: mutate(**{f"neg_{c.name}": -c for c in s})),
    )


def test_mul(df_int):
    assert_result_equal(
        df_int,
        lambda t: t >> (lambda s: mutate(**{f"mul_{c.name}_{d.name}": c * d for d in s for c in s})),
    )


def test_truediv(df_int):
    assert_result_equal(
        df_int,
        lambda t: t
        >> mutate(**{c.name: c.map({0: 1}, default=c) for c in t})
        >> (lambda s: mutate(**{f"div_{c.name}_{d.name}": c / d for d in s for c in s})),
    )


def test_floordiv(df_int):
    assert_result_equal(
        df_int,
        lambda t: t
        >> mutate(**{c.name: c.map({0: 1}, default=c) for c in t})
        >> (lambda s: mutate(**{f"div_{c.name}_{d.name}": c // d for d in s for c in s})),
    )


def test_mod(df_int):
    assert_result_equal(
        df_int,
        lambda t: t
        >> mutate(**{c.name: c.map({0: 1}, default=c) for c in t})
        >> (lambda s: mutate(**{f"mod_{c.name}_{d.name}": c % d for d in s for c in s})),
    )

    assert_result_equal(
        df_int,
        lambda t: t
        >> mutate(**{c.name: c.map({0: 1}, default=c) for c in t})
        >> (
            lambda s: mutate(**{f"mod_{c.name}_{d.name}": c % d for d in s for c in s})
            >> mutate(**{f"div_{c.name}_{d.name}": c // d for d in s for c in s})
        )
        >> summarize(
            **{
                f"div_plus_mod_{c.name}_{d.name}": (
                    (C[f"div_{c.name}_{d.name}"] * C[d.name] + C[f"mod_{c.name}_{d.name}"]) == C[c.name]
                ).all()
                for d in t
                for c in t
            }
        ),
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import contextlib
import warnings

import polars as pl
import pytest
from polars.testing import assert_frame_equal, assert_series_equal

from pydiverse.transform._internal.backend.targets import Polars
from pydiverse.transform._internal.errors import NonStandardWarning
from pydiverse.transform._internal.pipe.table import Table
from pydiverse.transform._internal.pipe.verbs import export, show_query


def assert_equal(left, right, check_dtypes=False, check_row_order=True):
    if isinstance(left, pl.Series):
        assert isinstance(right, pl.Series)
        assert_series_equal(left, right, check_names=False)
        return

    left_df = left >> export(Polars(lazy=False)) if isinstance(left, Table) else left
    right_df = right >> export(Polars(lazy=False)) if isinstance(right, Table) else right

    try:
        assert_frame_equal(
            left_df,
            right_df,
            check_column_order=False,
            check_row_order=check_row_order,
            check_dtypes=check_dtypes,
        )
    except AssertionError as e:
        print("First dataframe:")
        print(left_df)
        if isinstance(left, Table):
            left >> show_query()
        print()
        print("Second dataframe:")
        print(right_df)
        if isinstance(right, Table):
            right >> show_query()
        raise e


@contextlib.contextmanager
def catch_warnings():
    with warnings.catch_warnings(record=True) as record:
        warnings.simplefilter("always")
        yield record


def get_transform_warnings(
    warnings_: list[warnings.WarningMessage],
) -> list[warnings.WarningMessage]:
    return [w for w in warnings_ if issubclass(w.category, NonStandardWarning)]


def assert_result_equal(
    input_tables,
    pipe_factory,
    *,
    exception=None,
    check_row_order: bool = False,
    may_throw: bool = False,
    xfail_warnings: bool = True,
):
    if not isinstance(input_tables[0], tuple | list):
        input_tables = (input_tables,)
    x, y = zip(*input_tables, strict=True)

    if exception and not may_throw:
        with pytest.raises(exception):
            pipe_factory(*x) >> export(Polars(lazy=False))
        with pytest.raises(exception):
            pipe_factory(*y) >> export(Polars(lazy=False))
        return

    did_raise_warning = False
    warnings_summary = None
    try:
        with catch_warnings() as warnings_record:
            query_x = pipe_factory(*x)
            query_y = pipe_factory(*y)

            dfx: pl.DataFrame = (query_x >> export(Polars(lazy=False))).with_columns(
                pl.col(pl.Decimal).cast(pl.Float64)
            )
            dfy: pl.DataFrame = (query_y >> export(Polars(lazy=False))).with_columns(
                pl.col(pl.Decimal).cast(pl.Float64)
            )
            # TODO: for some reason, REALs on postgres and SQLite are converted to
            # Float64 instead of Float32.

            # TODO: after a join, cols containing only null values get type Null on
            # SQLite and Postgres. maybe we can fix this but for now we just ignore them
            assert dfx.columns == dfy.columns
            null_cols = set(dfx.select(pl.col(pl.Null)).columns) | set(dfy.select(pl.col(pl.Null)).columns)
            assert all(all(d.get_column(col).is_null().all() for col in null_cols) for d in (dfx, dfy))
            dfy = dfy.select(pl.all().exclude(null_cols))
            dfx = dfx.select(pl.all().exclude(null_cols))

        warnings_record = get_transform_warnings(warnings_record)
        if len(warnings_record):
            did_raise_warning = True
            warnings_summary = "\n".join({str(w.message) for w in warnings_record})

    except Exception as e:
        if may_throw:
            if exception is not None:
                if isinstance(exception, type):
                    exception = (exception,)
                if not isinstance(e, exception):
                    raise Exception(f"Raised the wrong type of exception: {type(e)} instead of {exception}.") from e
            # TODO: Replace with logger
            print(f"An exception was thrown:\n{e}")
            return
        else:
            raise e

    try:
        assert_frame_equal(dfx, dfy, check_row_order=check_row_order, check_exact=False, atol=1e-6)
    except Exception as e:
        if xfail_warnings and did_raise_warning:
            pytest.xfail(warnings_summary)

        print("First dataframe:")
        print(dfx)
        query_x >> show_query()
        print("")
        print("Second dataframe:")
        print(dfy)
        query_y >> show_query()
        print("")
        raise e



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import polars as pl
import pytest
from polars.testing import assert_series_equal

import pydiverse.transform as pdt
from pydiverse.transform._internal.errors import ColumnNotFoundError
from pydiverse.transform._internal.pipe.pipeable import Pipeable, inverse_partial
from pydiverse.transform.common import *


@pytest.fixture
def tbl1():
    return pdt.Table(pl.DataFrame({"col1": [0.1], "col2": [3.14]}))


@pytest.fixture
def tbl2():
    return pdt.Table(pl.DataFrame({"col1": [1], "col2": [2], "col3": [3]}))


class TestTable:
    def test_getattr(self, tbl1):
        assert tbl1.col1.name == "col1"
        assert tbl1.col2._ast == tbl1._ast

        with pytest.raises(ColumnNotFoundError, match="colXXX"):
            _ = tbl1.colXXX

    def test_getitem(self, tbl1):
        assert tbl1.col1._uuid == tbl1["col1"]._uuid
        assert tbl1.col2._uuid is tbl1["col2"]._uuid

    def test_iter(self, tbl1, tbl2):
        assert repr(list(tbl1)) == repr([tbl1.col1, tbl1.col2])
        assert repr(list(tbl2)) == repr([tbl2.col1, tbl2.col2, tbl2.col3])

        assert repr(list(tbl2 >> select(tbl2.col2))) == repr([tbl2.col2])

    def test_dir(self, tbl1):
        assert dir(tbl1) == ["col1", "col2"]
        assert dir(tbl1 >> mutate(x=tbl1.col1)) == ["col1", "col2", "x"]

    def test_contains(self, tbl1, tbl2):
        assert tbl1.col1 in tbl1
        assert tbl1.col2 in tbl1

        assert tbl1.col1 not in tbl2
        assert tbl1.col2 not in tbl2

        assert C.col1 in tbl1
        assert C.col2 in tbl1
        assert C.col3 not in tbl1

        assert C.col1 in tbl2
        assert C.col2 in tbl2
        assert C.col3 in tbl2
        assert C.col4 not in tbl2

        assert all(col in tbl1 for col in tbl1)
        assert all(col in tbl2 for col in tbl2)

    def test_dict_constructor(self):
        t = pdt.Table({"a": [3, 1, 4, 1, 5, 9, 4]})
        assert_series_equal(
            (t >> mutate(r=pdt.rank(arrange=t.a))).r.export(Polars()),
            pl.Series([3, 1, 4, 1, 6, 7, 4]),
            check_names=False,
        )


class TestDispatchers:
    def test_inverse_partial(self):
        def x(a, b, c):
            return (a, b, c)

        assert inverse_partial(x, 1, 2)(0) == (0, 1, 2)
        assert inverse_partial(x, 1, c=2)(0, c=3) == (0, 1, 3)

    def test_pipeable(self):
        @verb
        def add(v1, v2):
            return v1 + v2

        @verb
        def subtract(v1, v2):
            return v1 - v2

        with pytest.raises(TypeError):
            assert 1 >> add(2) == 3

        add_10 = add(5) >> add(5)
        assert isinstance(add_10, Pipeable)

        with pytest.raises(TypeError):
            assert 5 >> add_10 == 15



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy
import dataclasses
import re
import uuid
from collections.abc import Callable, Iterable
from typing import Literal
from uuid import UUID

from pydiverse.transform._internal.tree.ast import AstNode
from pydiverse.transform._internal.tree.col_expr import Col, ColExpr, Order


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class Verb(AstNode):
    child: AstNode

    def __post_init__(self):
        self.name = self.child.name

    def _unformatted_ast_repr(self, verb_depth: int, expr_depth: int, display_name_map):
        nd_repr = self._ast_node_repr(expr_depth, display_name_map)
        return (
            self.child._unformatted_ast_repr(verb_depth - 1, expr_depth, display_name_map) if verb_depth != 0 else "..."
        ) + f" >> {nd_repr}"

    def _clone(self) -> tuple["Verb", dict[AstNode, AstNode], dict[UUID, UUID]]:
        child, nd_map, uuid_map = self.child._clone()
        cloned = copy.copy(self)
        cloned.child = child

        cloned.map_col_nodes(
            lambda col: Col(
                col.name,
                # If the current ast is not in nd_map (happens after collect with
                # keep_col_refs=True), the node wasn't really present anyway.
                nd_map.get(col._ast, col._ast),
                uuid_map.get(col._uuid, col._uuid),
                col._dtype,
                col._ftype,
            )
            if isinstance(col, Col)
            else copy.copy(col)
        )
        nd_map[self] = cloned

        return cloned, nd_map, uuid_map

    def short_name(self):
        last_verb = self if not isinstance(self, SubqueryMarker) else self.child
        return (
            "?" if self.name is None else self.name
        ) + f" (last verb: {camel_to_snake(last_verb.__class__.__name__)})"

    def iter_subtree_postorder(self) -> Iterable[AstNode]:
        yield from self.child.iter_subtree_postorder()
        yield self

    def iter_subtree_preorder(self):
        yield self
        yield from self.child.iter_subtree_preorder()

    def iter_col_roots(self) -> Iterable[ColExpr]:
        return iter(())

    def iter_col_nodes(self) -> Iterable[ColExpr]:
        for col in self.iter_col_roots():
            yield from col.iter_subtree_postorder()

    def map_col_roots(self, g: Callable[[ColExpr], ColExpr]): ...

    def map_col_nodes(self, g: Callable[[ColExpr], ColExpr]):
        self.map_col_roots(lambda root: root.map_subtree(g))


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class Alias(Verb):
    uuid_map: dict[UUID, UUID] | None

    # TODO: currently we kinda misuse Alias for column reference transfer. Maybe it
    # would be nice to create a separate marker node for this to distinguish it from a
    # user-created alias.
    def _ast_node_repr(self, expr_depth: int, display_name_map) -> str:
        return "alias(" + (f"'{self.name}'" if self.name != self.child.name else "") + ")"

    def _clone(self) -> tuple[Verb, dict[AstNode, AstNode], dict[UUID, UUID]]:
        cloned, nd_map, uuid_map = Verb._clone(self)
        if self.uuid_map is not None:  # happens if and only if keep_col_refs=False
            assert set(self.uuid_map.keys()).issubset(uuid_map.keys())
            uuid_map = {
                self.uuid_map[old_uid]: new_uid for old_uid, new_uid in uuid_map.items() if old_uid in self.uuid_map
            }
            cloned.uuid_map = None
        return cloned, nd_map, uuid_map


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class Select(Verb):
    select: list[Col]

    def iter_col_roots(self) -> Iterable[ColExpr]:
        yield from self.select

    def map_col_roots(self, g: Callable[[ColExpr], ColExpr]):
        self.select = [g(col) for col in self.select]

    def _ast_node_repr(self, expr_depth: int, display_name_map) -> str:
        return "select(" + ", ".join(col._ast_repr(expr_depth, False, display_name_map) for col in self.select) + ")"


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class Rename(Verb):
    name_map: dict[str, str]

    def _ast_node_repr(self, expr_depth: int, display_name_map) -> str:
        return "rename({" + ", ".join(f"'{k}': '{v}'" for k, v in self.name_map.items()) + "})"


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class Mutate(Verb):
    names: list[str]
    values: list[ColExpr]
    uuids: list[UUID]

    def _ast_node_repr(self, expr_depth: int, display_name_map) -> str:
        return "mutate(" + (
            ", ".join(
                f"{k}={v._ast_repr(expr_depth, False, display_name_map)}"
                for k, v in zip(self.names, self.values, strict=True)
            )
            + ")"
        )

    def iter_col_roots(self) -> Iterable[ColExpr]:
        yield from self.values

    def map_col_roots(self, g: Callable[[ColExpr], ColExpr]):
        self.values = [g(val) for val in self.values]

    def _clone(self) -> tuple[Verb, dict[AstNode, AstNode], dict[UUID, UUID]]:
        cloned, nd_map, uuid_map = Verb._clone(self)
        assert isinstance(cloned, Mutate)
        cloned.uuids = [uuid.uuid1() for _ in self.names]
        uuid_map.update({old_uid: new_uid for old_uid, new_uid in zip(self.uuids, cloned.uuids, strict=True)})
        return cloned, nd_map, uuid_map


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class Filter(Verb):
    predicates: list[ColExpr]

    def _ast_node_repr(self, expr_depth: int, display_name_map) -> str:
        return (
            "filter(" + ", ".join(pred._ast_repr(expr_depth, False, display_name_map) for pred in self.predicates) + ")"
        )

    def iter_col_roots(self) -> Iterable[ColExpr]:
        yield from self.predicates

    def map_col_roots(self, g: Callable[[ColExpr], ColExpr]):
        self.predicates = [g(predicate) for predicate in self.predicates]


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class Summarize(Verb):
    names: list[str]
    values: list[ColExpr]
    uuids: list[UUID]

    def _ast_node_repr(self, expr_depth: int, display_name_map: dict) -> str:
        return "summarize(" + (
            ", ".join(
                f"{k}={v._ast_repr(expr_depth, False, display_name_map)}"
                for k, v in zip(self.names, self.values, strict=True)
            )
            + ")"
        )

    def iter_col_roots(self) -> Iterable[ColExpr]:
        yield from self.values

    def map_col_roots(self, g: Callable[[ColExpr], ColExpr]):
        self.values = [g(val) for val in self.values]

    def _clone(self) -> tuple[Verb, dict[AstNode, AstNode], dict[UUID, UUID]]:
        cloned, nd_map, uuid_map = Verb._clone(self)
        assert isinstance(cloned, Summarize)
        cloned.uuids = [uuid.uuid1() for _ in self.names]
        uuid_map.update({old_uid: new_uid for old_uid, new_uid in zip(self.uuids, cloned.uuids, strict=True)})
        return cloned, nd_map, uuid_map


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class Arrange(Verb):
    order_by: list[Order]

    def _ast_node_repr(self, expr_depth: int, display_name_map) -> str:
        return "arrange(" + ", ".join(ord._ast_repr(expr_depth, False, display_name_map) for ord in self.order_by) + ")"

    def iter_col_roots(self) -> Iterable[ColExpr]:
        yield from (ord.order_by for ord in self.order_by)

    def map_col_roots(self, g: Callable[[ColExpr], ColExpr]):
        self.order_by = [Order(g(ord.order_by), ord.descending, ord.nulls_last) for ord in self.order_by]


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class SliceHead(Verb):
    n: int
    offset: int

    def _ast_node_repr(self, expr_depth: int, display_name_map) -> str:
        return f"slice_head(n={self.n}, offset={self.offset})"


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class GroupBy(Verb):
    group_by: list[Col]
    add: bool

    def _ast_node_repr(self, expr_depth: int, display_name_map) -> str:
        return (
            "group_by(" + ", ".join(col._ast_repr(expr_depth, False, display_name_map) for col in self.group_by) + ")"
        )

    def iter_col_roots(self) -> Iterable[ColExpr]:
        yield from self.group_by

    def map_col_roots(self, g: Callable[[ColExpr], ColExpr]):
        self.group_by = [g(col) for col in self.group_by]


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class Ungroup(Verb):
    def _ast_node_repr(self, expr_depth, display_name_map):
        return "ungroup()"


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class Join(Verb):
    right: AstNode
    on: ColExpr
    how: Literal["inner", "left", "full"]
    validate: Literal["1:1", "1:m", "m:1", "m:m"]

    def _unformatted_ast_repr(self, verb_depth: int, expr_depth: int, display_name_map) -> str:
        return (
            self.child._unformatted_ast_repr(verb_depth - 1, expr_depth, display_name_map) if verb_depth != 0 else "..."
        ) + (
            ">> join("
            + (
                self.right._unformatted_ast_repr(verb_depth - 1, expr_depth, display_name_map)
                if verb_depth != 0
                else self.right._unformatted_ast_repr(0, expr_depth, display_name_map)
            )
            + ", "
            + f"how='{self.how}', "
            + f"on={self.on._ast_repr(expr_depth, False, display_name_map)}, "
            + f"validate='{self.validate}')"
        )

    def _clone(self) -> tuple["Join", dict[AstNode, AstNode], dict[UUID, UUID]]:
        child, nd_map, uuid_map = self.child._clone()
        right_child, right_nd_map, right_uuid_map = self.right._clone()
        nd_map.update(right_nd_map)
        uuid_map.update(right_uuid_map)

        cloned = copy.copy(self)
        cloned.child = child
        cloned.right = right_child
        cloned.on = self.on.map_subtree(
            lambda col: Col(
                col.name,
                nd_map.get(col._ast, col._ast),
                uuid_map[col._uuid],
                col._dtype,
                col._ftype,
            )
            if isinstance(col, Col)
            else copy.copy(col)
        )

        nd_map[self] = cloned
        return cloned, nd_map, uuid_map

    def iter_subtree_postorder(self) -> Iterable[AstNode]:
        yield from self.child.iter_subtree_postorder()
        yield from self.right.iter_subtree_postorder()
        yield self

    def iter_subtree_preorder(self):
        yield self
        yield from self.child.iter_subtree_preorder()
        yield from self.right.iter_subtree_preorder()

    def iter_col_roots(self) -> Iterable[ColExpr]:
        yield self.on

    def map_col_roots(self, g: Callable[[ColExpr], ColExpr]):
        self.on = g(self.on)


@dataclasses.dataclass(eq=False, slots=True, repr=False)
class Union(Verb):
    right: AstNode
    distinct: bool

    def _unformatted_ast_repr(self, verb_depth: int, expr_depth: int, display_name_map) -> str:
        return (
            self.child._unformatted_ast_repr(verb_depth - 1, expr_depth, display_name_map) if verb_depth != 0 else "..."
        ) + (
            ">> union("
            + (
                self.right._unformatted_ast_repr(verb_depth - 1, expr_depth, display_name_map)
                if verb_depth != 0
                else self.right._unformatted_ast_repr(0, expr_depth, display_name_map)
            )
            + f", distinct={self.distinct})"
        )

    def _clone(self) -> tuple["Union", dict[AstNode, AstNode], dict[UUID, UUID]]:
        child, nd_map, uuid_map = self.child._clone()
        right_child, right_nd_map, right_uuid_map = self.right._clone()
        nd_map.update(right_nd_map)
        uuid_map.update(right_uuid_map)

        cloned = copy.copy(self)
        cloned.child = child
        cloned.right = right_child

        nd_map[self] = cloned
        return cloned, nd_map, uuid_map

    def iter_subtree_postorder(self) -> Iterable[AstNode]:
        yield from self.child.iter_subtree_postorder()
        yield from self.right.iter_subtree_postorder()
        yield self

    def iter_subtree_preorder(self):
        yield self
        yield from self.child.iter_subtree_preorder()
        yield from self.right.iter_subtree_preorder()

    def iter_col_roots(self) -> Iterable[ColExpr]:
        return iter(())

    def map_col_roots(self, g: Callable[[ColExpr], ColExpr]):
        pass


class SubqueryMarker(Verb):
    def _ast_node_repr(self, expr_depth, display_name_map):
        return "subquery_marker"


CAMEL_TO_SNAKE = re.compile(r"(?<!^)(?=[A-Z])")


def camel_to_snake(s: str) -> str:
    return CAMEL_TO_SNAKE.sub("_", s).lower()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform.extended import *
from tests.util.assertion import assert_result_equal


def test_or(df_bool):
    assert_result_equal(df_bool, lambda t: t >> mutate(y=t.a | t.b))


def test_and(df_bool):
    assert_result_equal(df_bool, lambda t: t >> mutate(y=t.a & t.b))


def test_xor(df_bool):
    assert_result_equal(df_bool, lambda t: t >> mutate(y=t.a ^ t.b))


def test_invert(df_bool):
    assert_result_equal(df_bool, lambda t: t >> mutate(y=~t.a))


def test_any_all(df_bool):
    assert_result_equal(df_bool, lambda t: t >> group_by(t.b) >> summarize(y=t.a.any()))
    assert_result_equal(df_bool, lambda t: t >> group_by(t.b) >> summarize(y=t.a.all()))


def test_bool_comparison(df_bool):
    assert_result_equal(df_bool, lambda t: t >> mutate(y=t.a < t.b))
    assert_result_equal(df_bool, lambda t: t >> mutate(y=t.a > t.b))
    assert_result_equal(df_bool, lambda t: t >> mutate(y=t.a <= t.b))
    assert_result_equal(df_bool, lambda t: t >> mutate(y=t.a >= t.b))


def test_bool_min_max(df_bool):
    assert_result_equal(df_bool, lambda t: t >> group_by(t.b) >> summarize(y=t.a.min()))
    assert_result_equal(df_bool, lambda t: t >> group_by(t.b) >> summarize(y=t.a.max()))


def test_bool_sum(df_bool):
    assert_result_equal(df_bool, lambda t: t >> mutate(y=t.a + t.b))
    assert_result_equal(df_bool, lambda t: t >> group_by(t.b) >> summarize(y=t.a.sum()))
    assert_result_equal(df_bool, lambda t: t >> group_by(t.a) >> summarize(y=t.b.sum()))



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.ops.op import Operator
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import S


class Marker(Operator):
    def __init__(self, name: str, doc: str = ""):
        super().__init__(
            name,
            Signature(S, return_type=S),
            doc=doc
            + """
Can only be used in expressions given to the `arrange` verb or as as an
`arrange` keyword argument.
""",
        )


nulls_first = Marker(
    "nulls_first",
    doc="""
Specifies that nulls are placed at the beginning of the ordering.

This does not mean that nulls are considered to be `less` than any other
element. I.e. if both `nulls_first` and `descending` are given, nulls will still
be placed at the beginning.

If neither `nulls_first` nor `nulls_last` is specified, the position of nulls is
backend-dependent.
""",
)

nulls_last = Marker(
    "nulls_last",
    doc="""
Specifies that nulls are placed at the end of the ordering.

This does not mean that nulls are considered to be `greater` than any other
element. I.e. if both `nulls_last` and `descending` are given, nulls will still
be placed at the end.

If neither `nulls_first` nor `nulls_last` is specified, the position of nulls is
backend-dependent.
""",
)

ascending = Marker(
    "ascending",
    doc="""
The default ordering.
""",
)

descending = Marker(
    "descending",
    doc="""
Reverses the default ordering.
""",
)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.pipe.verbs import (
    select,
)
from tests.util import assert_result_equal


def test_simple_select(df1):
    assert_result_equal(df1, lambda t: t >> select(t.col1))
    assert_result_equal(df1, lambda t: t >> select(t.col2))
    assert_result_equal(df1, lambda t: t >> select("col1"))


def test_reorder(df1):
    assert_result_equal(df1, lambda t: t >> select(t.col2, t.col1))



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.ops.op import Operator
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import (
    COMPARABLE,
    NUMERIC,
    Bool,
    Duration,
    S,
    String,
)


class Horizontal(Operator):
    def __init__(self, name: str, *signatures: Signature, doc: str = ""):
        super().__init__(
            name,
            *signatures,
            param_names=["arg", "args"],
            generate_expr_method=False,
            doc=doc,
        )


horizontal_max = Horizontal(
    "max",
    *(Signature(dtype, dtype, ..., return_type=dtype) for dtype in COMPARABLE),
    doc="""
The maximum of the given columns.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": [5, None, 435, -1, 8, None],
...         "b": [-45, None, 6, 23, -1, 0],
...         "c": [10, None, 2, None, -53, 3],
...     }
... )
>>> t >> mutate(x=pdt.max(t.a, t.b, t.c)) >> show()
Table <unnamed>, backend: PolarsImpl
shape: (6, 4)
┌──────┬──────┬──────┬──────┐
│ a    ┆ b    ┆ c    ┆ x    │
│ ---  ┆ ---  ┆ ---  ┆ ---  │
│ i64  ┆ i64  ┆ i64  ┆ i64  │
╞══════╪══════╪══════╪══════╡
│ 5    ┆ -45  ┆ 10   ┆ 10   │
│ null ┆ null ┆ null ┆ null │
│ 435  ┆ 6    ┆ 2    ┆ 435  │
│ -1   ┆ 23   ┆ null ┆ 23   │
│ 8    ┆ -1   ┆ -53  ┆ 8    │
│ null ┆ 0    ┆ 3    ┆ 3    │
└──────┴──────┴──────┴──────┘
""",
)

horizontal_min = Horizontal(
    "min",
    *(Signature(dtype, dtype, ..., return_type=dtype) for dtype in COMPARABLE),
    doc="""
The minimum of the given columns.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": [5, None, 435, -1, 8, None],
...         "b": [-45, None, 6, 23, -1, 0],
...         "c": [10, None, 2, None, -53, 3],
...     }
... )
>>> t >> mutate(x=pdt.min(t.a, t.b, t.c)) >> show()
Table <unnamed>, backend: PolarsImpl
shape: (6, 4)
┌──────┬──────┬──────┬──────┐
│ a    ┆ b    ┆ c    ┆ x    │
│ ---  ┆ ---  ┆ ---  ┆ ---  │
│ i64  ┆ i64  ┆ i64  ┆ i64  │
╞══════╪══════╪══════╪══════╡
│ 5    ┆ -45  ┆ 10   ┆ -45  │
│ null ┆ null ┆ null ┆ null │
│ 435  ┆ 6    ┆ 2    ┆ 2    │
│ -1   ┆ 23   ┆ null ┆ -1   │
│ 8    ┆ -1   ┆ -53  ┆ -53  │
│ null ┆ 0    ┆ 3    ┆ 0    │
└──────┴──────┴──────┴──────┘
""",
)

coalesce = Horizontal(
    "coalesce",
    Signature(S, S, ..., return_type=S),
    doc="""
Returns the first non-null value among the given.

:param arg:
    The first value.

:param args:
    Further values. All must have the same type.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": [5, None, 435, -1, 8, None],
...         "b": [-45, None, 6, 23, 1, 0],
...         "c": [10, 2, None, None, None, None],
...     }
... )
>>> (
...     t
...     >> mutate(
...         x=pdt.coalesce(t.a, t.b, t.c),
...         y=pdt.coalesce(t.c, t.b, t.a),
...     )
...     >> show()
... )
Table <unnamed>, backend: PolarsImpl
shape: (6, 5)
┌──────┬──────┬──────┬─────┬─────┐
│ a    ┆ b    ┆ c    ┆ x   ┆ y   │
│ ---  ┆ ---  ┆ ---  ┆ --- ┆ --- │
│ i64  ┆ i64  ┆ i64  ┆ i64 ┆ i64 │
╞══════╪══════╪══════╪═════╪═════╡
│ 5    ┆ -45  ┆ 10   ┆ 5   ┆ 10  │
│ null ┆ null ┆ 2    ┆ 2   ┆ 2   │
│ 435  ┆ 6    ┆ null ┆ 435 ┆ 6   │
│ -1   ┆ 23   ┆ null ┆ -1  ┆ 23  │
│ 8    ┆ 1    ┆ null ┆ 8   ┆ 1   │
│ null ┆ 0    ┆ null ┆ 0   ┆ 0   │
└──────┴──────┴──────┴─────┴─────┘
""",
)

horizontal_any = Horizontal("any", Signature(Bool(), Bool(), ..., return_type=Bool()))

horizontal_all = Horizontal("all", Signature(Bool(), Bool(), ..., return_type=Bool()))

horizontal_sum = Horizontal(
    "sum",
    *(Signature(dtype, dtype, ..., return_type=dtype) for dtype in NUMERIC),
    Signature(String(), String(), ..., return_type=String()),
    Signature(Duration(), Duration(), ..., return_type=Duration()),
)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.ops.op import Operator
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import Date, Datetime, Duration, Int


class DatetimeExtract(Operator):
    def __init__(self, name: str, doc: str | None = None):
        super().__init__(
            name,
            Signature(Datetime(), return_type=Int()),
            doc=doc if doc is not None else f"Extracts the {name[3:]} component.",
        )


class DateExtract(Operator):
    def __init__(self, name: str, doc: str | None = None):
        super().__init__(
            name,
            Signature(Date(), return_type=Int()),
            Signature(Datetime(), return_type=Int()),
            doc=doc if doc is not None else f"Extracts the {name[3:]} component.",
        )


dt_year = DateExtract("dt.year", doc="The year component of the date or datetime.")

dt_month = DateExtract("dt.month", doc="The month component of the date or datetime.")

dt_day = DateExtract("dt.day", doc="The day component of the date or datetime.")

dt_hour = DatetimeExtract("dt.hour", doc="The hour component of the datetime.")

dt_minute = DatetimeExtract("dt.minute", doc="The minute component of the datetime.")

dt_second = DatetimeExtract("dt.second", doc="The second component of the datetime.")

dt_millisecond = DatetimeExtract(
    "dt.millisecond",
    doc="""
The microsecond component of the datetime int-divided by 1000 (see polars).
""",
)

dt_microsecond = DatetimeExtract("dt.microsecond", doc="The microsecond component of the datetime.")

dt_day_of_week = DateExtract(
    "dt.day_of_week",
    doc="""
The number of the current weekday.

This is one-based, so Monday is 1 and Sunday is 7.
""",
)

dt_day_of_year = DateExtract(
    "dt.day_of_year",
    doc="""
The number of days since the beginning of the year.

This is one-based, so it returns 1 for the 1st of January.
""",
)


class DurationToUnit(Operator):
    def __init__(self, name: str, doc: str = ""):
        super().__init__(name, Signature(Duration(), return_type=Int()), doc=doc)


dur_days = DurationToUnit("dur.days")

dur_hours = DurationToUnit("dur.hours")

dur_minutes = DurationToUnit("dur.minutes")

dur_seconds = DurationToUnit("dur.seconds")

dur_milliseconds = DurationToUnit("dur.milliseconds")

dur_microseconds = DurationToUnit("dur.microseconds")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.ops.op import Operator
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import COMPARABLE, Bool, Const, S

equal = Operator("__eq__", Signature(S, S, return_type=Bool()), doc="Equality comparison ==")

not_equal = Operator("__ne__", Signature(S, S, return_type=Bool()), doc="Non-equality comparison !=")


less_than = Operator(
    "__lt__",
    *(Signature(t, t, return_type=Bool()) for t in COMPARABLE),
    doc="Less than comparison <",
)

less_equal = Operator(
    "__le__",
    *(Signature(t, t, return_type=Bool()) for t in COMPARABLE),
    doc="Less than or equal to comparison <=",
)

greater_than = Operator(
    "__gt__",
    *(Signature(t, t, return_type=Bool()) for t in COMPARABLE),
    doc="Greater than comparison >",
)

greater_equal = Operator(
    "__ge__",
    *(Signature(t, t, return_type=Bool()) for t in COMPARABLE),
    doc="Greater than or equal to comparison >=",
)

is_null = Operator(
    "is_null",
    Signature(S, return_type=Bool()),
    doc="Indicates whether the value is null.",
)

is_not_null = Operator(
    "is_not_null",
    Signature(S, return_type=Bool()),
    doc="Indicates whether the value is not null.",
)

fill_null = Operator(
    "fill_null",
    Signature(S, S, return_type=S),
    doc="Replaces every null by the given value.",
)

is_in = Operator(
    "is_in",
    Signature(S, S, ..., return_type=Bool()),
    doc="""
Whether the value equals one of the given.

Note
----
The expression ``t.c.is_in(a1, a2, ...)`` is equivalent to
``(t.c == a1) | (t.c == a2) | ...``, so passing null to ``is_in`` will result in
null. To compare for equality with null, use
:doc:`pydiverse.transform.ColExpr.is_null`.
""",
)

clip = Operator(
    "clip",
    *(Signature(t, Const(t), Const(t), return_type=t) for t in COMPARABLE),
    param_names=["self", "lower_bound", "upper_bound"],
    doc="""
Replaces values outside `[lower_bound, upper_bound]` with the closer boundary
value. If the input is not null, this is equivalent to `pdt.max(pdt.min(self,
upper_bound), lower_bound)`.
""",
)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pydiverse.transform as pdt
from pydiverse.transform import C
from pydiverse.transform._internal.errors import DataTypeError, FunctionTypeError
from pydiverse.transform._internal.pipe.verbs import (
    group_by,
    mutate,
    summarize,
)
from tests.util import assert_result_equal


def test_mutate_case_ewise(df4):
    assert_result_equal(
        df4,
        lambda t: t >> mutate(x=C.col1.map({0: 1, (1, 2): 2}), y=C.col1.map({0: 0, 1: None}, default=10.4)),
    )

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x=pdt.when(C.col1 == C.col2).then(1).when(C.col2 == C.col3).then(2).otherwise(C.col1 + C.col2),
        ),
    )

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x=t.col1.map({0: 1}),
            y=t.col4.map({2: 2}, default=-1),
            z=t.col2.map({4: 3}, default=t.col1),
            w=t.col3.map({2: 2, 1: 5}, default=pdt.lit(None)),
        ),
    )


def test_mutate_case_window(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x=pdt.when(C.col1.max() == 1)
            .then(1)
            .when(C.col1.max() == 2)
            .then(2)
            .when(C.col1.max() == 3)
            .then(3)
            .when(C.col1.max() == 4)
            .then(4)
        ),
    )

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            u=C.col1.shift(1, 1729, arrange=[t.col3.descending().nulls_last(), t.col4.nulls_last()]),
            x=C.col1.shift(1, 0, arrange=[C.col4.nulls_first()]).map(
                {
                    1: C.col2.shift(1, -1, arrange=[C.col2.nulls_last(), C.col4.nulls_first()]),
                    2: C.col3.shift(2, -2, arrange=[C.col3.nulls_last(), C.col4.nulls_last()]),
                }
            ),
        ),
    )

    # Can't nest window in window
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x=C.col1.shift(1, 0, arrange=[C.col4.nulls_last()])
            .map(
                {
                    1: 2,
                    2: 3,
                }
            )
            .shift(1, -1, arrange=[C.col4.descending().nulls_first()])
        ),
        may_throw=True,
    )


def test_summarize_case(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(
            C.col1,
        )
        >> summarize(
            x=C.col2.max().map(
                {
                    0: C.col1.min(),
                    1: C.col2.mean() + 0.5,
                    2: 2,
                }
            ),
            y=pdt.when(C.col2.max() > 2).then(1).when(C.col2.max() < 2).then(C.col2.min()).otherwise(C.col3.mean()),
        ),
    )


def test_invalid_value_dtype(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x=C.col1.map(
                {
                    0: "a",
                    1: 1.1,
                }
            )
        ),
        exception=DataTypeError,
    )


def test_invalid_ftype(df1):
    assert_result_equal(
        df1,
        lambda t: t
        >> summarize(
            x=pdt.rank(arrange=[C.col1]).map(
                {
                    1: C.col1.max(),
                },
                default=None,
            )
        ),
        exception=FunctionTypeError,
    )

    assert_result_equal(
        df1,
        lambda t: t >> summarize(x=pdt.when(pdt.rank(arrange=[C.col1]) == 1).then(1).otherwise(None)),
        exception=FunctionTypeError,
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from ._internal.backend.targets import (
    Dict,
    DictOfLists,
    DuckDb,
    ListOfDicts,
    Pandas,
    Polars,
    Scalar,
    SqlAlchemy,
)
from ._internal.pipe.aligned import aligned, eval_aligned
from ._internal.pipe.cache import transfer_col_references
from ._internal.pipe.pipeable import verb
from ._internal.pipe.verbs import (
    arrange,
    ast_repr,
    columns,
    cross_join,
    drop,
    full_join,
    group_by,
    inner_join,
    join,
    left_join,
    mutate,
    name,
    rename,
    select,
    slice_head,
    summarize,
    ungroup,
    union,
)
from .base import *  # noqa: F403
from .base import __all__ as __base

__all__ = __base + [
    "verb",
    "aligned",
    "eval_aligned",
    "arrange",
    "ast_repr",
    "columns",
    "drop",
    "group_by",
    "name",
    "join",
    "inner_join",
    "left_join",
    "full_join",
    "cross_join",
    "union",
    "mutate",
    "rename",
    "select",
    "slice_head",
    "summarize",
    "ungroup",
    "DuckDb",
    "SqlAlchemy",
    "Polars",
    "Pandas",
    "Scalar",
    "Dict",
    "DictOfLists",
    "ListOfDicts",
    "transfer_col_references",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pydiverse.transform as pdt
from pydiverse.transform.errors import ColumnNotFoundError
from pydiverse.transform.extended import *
from tests.util import assert_result_equal


def test_simple(df3):
    assert_result_equal(df3, lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(1))
    assert_result_equal(df3, lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(10))
    assert_result_equal(df3, lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(100))

    assert_result_equal(df3, lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(1, offset=8))
    assert_result_equal(df3, lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(10, offset=8))
    assert_result_equal(df3, lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(100, offset=8))

    assert_result_equal(df3, lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(1, offset=100))
    assert_result_equal(df3, lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(10, offset=100))
    assert_result_equal(df3, lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(100, offset=100))


def test_chained(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> arrange(*list(t)[0:-1])
        >> slice_head(1)
        >> alias()
        >> (lambda s: arrange(*list(s)[0:-1]) >> slice_head(1)),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> arrange(*list(t)[0:-1])
        >> slice_head(10)
        >> alias()
        >> (lambda s: arrange(*list(s)[0:-1]) >> slice_head(5)),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> arrange(*list(t)[0:-1])
        >> slice_head(2, offset=5)
        >> alias()
        >> (lambda s: arrange(*list(s)[0:-1]) >> slice_head(2, offset=1)),
    )


def test_with_mutate(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(a=C.col1 * 2)
        >> arrange(*list(t)[0:-1])
        >> slice_head(4, offset=2)
        >> mutate(b=C.col2 + C.a),
    )


def test_with_join(df1, df2):
    assert_result_equal(
        (df1, df2),
        lambda t, u: t >> arrange(*t) >> slice_head(3) >> alias(keep_col_refs=True) >> left_join(u, t.col1 == u.col1),
        check_row_order=False,
    )

    assert_result_equal(
        (df1, df2),
        lambda t, u: t >> left_join(u >> arrange(*t) >> slice_head(2, offset=1), t.col1 == u.col1),
        check_row_order=False,
        exception=ColumnNotFoundError,
        may_throw=True,
    )


def test_with_filter(df3):
    assert_result_equal(
        df3,
        lambda t: t >> filter(t.col4 % 2 == 0) >> arrange(*list(t)[0:-1]) >> slice_head(4, offset=2),
    )

    assert_result_equal(
        df3,
        lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(4, offset=2) >> alias() >> filter(C.col1 == 1),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> filter(t.col4 % 2 == 0)
        >> arrange(*list(t)[0:-1])
        >> slice_head(4, offset=2)
        >> alias()
        >> filter(C.col1 == 1),
    )


def test_with_arrange(df3):
    assert_result_equal(
        df3,
        lambda t: t >> mutate(x=t.col4 - (t.col1 * t.col2)) >> arrange(C.x, *list(t)[0:-1]) >> slice_head(4, offset=2),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(x=(t.col1 * t.col2))
        >> arrange(*list(t)[0:-1])
        >> slice_head(4)
        >> alias()
        >> (lambda _: arrange(-C.x, C.col5)),
    )


def test_with_group_by(df3):
    assert_result_equal(
        df3,
        lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(1) >> alias() >> group_by(C.col1) >> mutate(x=pdt.count()),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> arrange(C.col1, *list(t)[0:-1])
        >> slice_head(6, offset=1)
        >> rename({"col2": "c"})
        >> alias()
        >> group_by(C.col1)
        >> mutate(x=C.col4.mean())
        >> select(C.x),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(key=C.col4 % (C.col3 + 1))
        >> arrange(C.key, *list(t)[0:-1])
        >> slice_head(4)
        >> alias()
        >> group_by(C.key)
        >> summarize(x=pdt.count()),
    )


def test_with_summarize(df3):
    assert_result_equal(
        df3,
        lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(4) >> alias() >> summarize(count=pdt.count()),
    )

    assert_result_equal(
        df3,
        lambda t: t >> arrange(*list(t)[0:-1]) >> slice_head(4) >> alias() >> summarize(c3_mean=C.col3.mean()),
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.pipe.verbs import (
    rename,
)
from tests.util import assert_result_equal


def test_noop(df3):
    assert_result_equal(df3, lambda t: t >> rename({}))
    assert_result_equal(df3, lambda t: t >> rename({"col1": "col1"}))


def test_simple(df3):
    assert_result_equal(df3, lambda t: t >> rename({"col1": "X"}))
    assert_result_equal(df3, lambda t: t >> rename({"col2": "Y"}))
    assert_result_equal(df3, lambda t: t >> rename({"col1": "A", "col2": "B"}))
    assert_result_equal(df3, lambda t: t >> rename({"col2": "B", "col1": "A"}))


def test_chained(df3):
    assert_result_equal(df3, lambda t: t >> rename({"col1": "X"}) >> rename({"X": "Y"}))
    assert_result_equal(df3, lambda t: t >> rename({"col1": "X"}) >> rename({"X": "col1"}))
    assert_result_equal(
        df3,
        lambda t: t >> rename({"col1": "1", "col2": "2"}) >> rename({"1": "col1", "2": "col2"}),
    )


def test_complex(df3):
    assert_result_equal(
        df3,
        lambda t: t >> rename({"col1": "col2", "col2": "col3", "col3": "col1"}),
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any
from uuid import UUID

import duckdb_engine.datatypes as duckdb_types
import polars as pl
import sqlalchemy as sqa
from sqlalchemy.dialects.postgresql import aggregate_order_by
from sqlalchemy.sql.type_api import TypeEngine as TypeEngine

from pydiverse.common import Decimal, Dtype, Int8, Int64, UInt8, UInt16, UInt32, UInt64
from pydiverse.transform._internal.backend import sql
from pydiverse.transform._internal.backend.sql import SqlImpl
from pydiverse.transform._internal.backend.targets import Polars, Target
from pydiverse.transform._internal.ops import ops
from pydiverse.transform._internal.tree import types
from pydiverse.transform._internal.tree.ast import AstNode
from pydiverse.transform._internal.tree.col_expr import Cast, LiteralCol


class DuckDbImpl(SqlImpl):
    backend_name = "duckdb"

    @classmethod
    def export(
        cls,
        nd: AstNode,
        target: Target,
        *,
        schema_overrides: dict[UUID, Any],
    ):
        if isinstance(target, Polars):
            engine = sql.get_engine(nd)
            with engine.connect() as conn:
                return pl.read_database(
                    DuckDbImpl.build_query(nd),
                    connection=conn,
                    schema_overrides=schema_overrides,
                )
        return SqlImpl.export(nd, target, schema_overrides=schema_overrides)

    @classmethod
    def compile_cast(cls, cast: Cast, sqa_col: dict[str, sqa.Label]) -> Cast:
        if cast.val.dtype().is_float() and cast.target_type.is_int():
            return cls.cast_compiled(cast, sqa.func.trunc(cls.compile_col_expr(cast.val, sqa_col)))

        return super().compile_cast(cast, sqa_col)

    @classmethod
    def compile_lit(cls, lit: LiteralCol) -> sqa.ColumnElement:
        if types.without_const(lit.dtype()) == Int64():
            return sqa.cast(lit.val, sqa.BigInteger)
        return super().compile_lit(lit)

    @classmethod
    def fix_fn_types(cls, fn: sql.ColFn, val: sqa.ColumnElement, *args: sqa.ColumnElement) -> sqa.ColumnElement:
        if fn.op in (ops.sum, ops.cum_sum):
            return sqa.cast(val, type_=args[0].type)
        return val

    @classmethod
    def compile_ordered_aggregation(
        cls, *args: sqa.ColumnElement, order_by: list[sqa.UnaryExpression], impl
    ) -> sqa.ColumnElement:
        return impl(*args[:-1], aggregate_order_by(args[-1], *order_by))

    @classmethod
    def sqa_type(cls, pdt_type: Dtype):
        if isinstance(pdt_type, Int8):
            return duckdb_types.TinyInteger()
        if isinstance(pdt_type, UInt8):
            return duckdb_types.UTinyInteger()
        if isinstance(pdt_type, UInt16):
            return duckdb_types.UInt16()
        if isinstance(pdt_type, UInt32):
            return duckdb_types.UInt32()
        if isinstance(pdt_type, UInt64):
            return duckdb_types.UInt64()
        if isinstance(pdt_type, Decimal):
            return sqa.DECIMAL(35, 10)

        return super().sqa_type(pdt_type)


with DuckDbImpl.impl_store.impl_manager as impl:

    @impl(ops.floordiv)
    def _floordiv(lhs, rhs):
        return sqa.func.divide(lhs, rhs)

    @impl(ops.is_inf)
    def _is_inf(x):
        return sqa.func.isinf(x)

    @impl(ops.is_not_inf)
    def _is_not_inf(x):
        return sqa.func.isfinite(x)

    @impl(ops.is_nan)
    def _is_nan(x):
        return sqa.func.isnan(x)

    @impl(ops.is_not_nan)
    def _is_not_nan(x):
        return ~sqa.func.isnan(x)

    @impl(ops.str_join)
    def _str_join(x, delim):
        return sqa.func.string_agg(x, delim)

    @impl(ops.list_agg)
    def _list_agg(x):
        return sqa.func.array_agg(x)

    @impl(ops.dt_day_of_week)
    def _day_of_week(x):
        return (sqa.extract("dow", x) + 6) % sqa.literal_column("7") + 1

    @impl(ops.str_contains)
    def _str_contains(x, pattern, allow_regex, true_if_regex_unsupported):
        if not allow_regex:
            return x.contains(pattern, autoescape=True)
        return sqa.func.regexp_matches(x, pattern)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.common import (
    Bool,
    Date,
    Datetime,
    Decimal,
    Dtype,
    Enum,
    Float,
    Float32,
    Float64,
    Int,
    Int8,
    Int16,
    Int32,
    Int64,
    List,
    String,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
)

__all__ = [
    "Dtype",
    "Bool",
    "Date",
    "Datetime",
    "Decimal",
    "Float",
    "Float32",
    "Float64",
    "Int",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "String",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
    "Dtype",
    "List",
    "Enum",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import inspect
from collections.abc import Callable, Sequence

from pydiverse.transform._internal.ops.op import Operator
from pydiverse.transform._internal.ops.signature import SignatureTrie
from pydiverse.transform._internal.tree.types import Dtype


class ImplStore:
    __slots__ = ("impl_trie", "default_impl", "impl_manager")

    impl_trie: dict[Operator, SignatureTrie]
    default_impl: dict[Operator, Callable | None]
    impl_manager: "ImplContextManager"

    def __init__(self) -> None:
        self.impl_trie = dict()
        self.default_impl = dict()
        self.impl_manager = ImplContextManager(self)

    def add_impl(
        self,
        op: Operator,
        sig: Sequence[Dtype] | None,
        is_vararg: bool,
        f: Callable,
    ) -> None:
        if sig is None:
            assert op not in self.default_impl
            self.default_impl[op] = f
        else:
            if op not in self.impl_trie:
                self.impl_trie[op] = SignatureTrie()
            self.impl_trie[op].insert(sig, f, is_vararg)

    def get_impl(self, op: Operator, sig: Sequence[Dtype]) -> Callable | None:
        best_match = None

        if (trie := self.impl_trie.get(op)) is not None:
            trie_match = trie.best_match(sig)
            if trie_match is not None:
                best_match = trie_match[1]
        if best_match is None:
            best_match = self.default_impl.get(op)

        if best_match is None:
            return None

        # filter out only those kwargs that the impl wants
        impl_kwargs = {
            name
            for name, param in inspect.signature(best_match).parameters.items()
            if param.kind == inspect.Parameter.KEYWORD_ONLY
        }

        return lambda *args, **kwargs: best_match(
            *args,
            **{kwarg: val for kwarg, val in kwargs.items() if kwarg in impl_kwargs},
            **({"_sig": sig} if "_sig" in impl_kwargs else {}),
        )


@dataclasses.dataclass(slots=True)
class ImplContextManager:
    impl_store: ImplStore

    def __enter__(self):
        return self

    def __exit__(self, *args): ...

    def __call__(self, op: Operator, *sig: Dtype) -> Callable:
        assert isinstance(op, Operator)

        is_vararg = len(sig) > 1 and sig[-1] is Ellipsis
        if is_vararg:
            sig = sig[:-1]

        if not all(isinstance(dtype, Dtype) for dtype in sig):
            raise TypeError(
                "The argument types must have type `Dtype`.\n"
                "hint: Maybe you forgot the parentheses and passed in the class "
                "instead of an instance? (E.g. it must be `Int()` instead of `Int`.)"
            )

        def f(g):
            self.impl_store.add_impl(op, None if len(sig) == 0 else sig, is_vararg, g)
            return g

        return f



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from .assertion import assert_equal, assert_result_equal

__all__ = [
    "assert_equal",
    "assert_result_equal",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from .aggregation import *
from .arithmetic import *
from .comparison import *
from .datetime import *
from .horizontal import *
from .list import *
from .logical import *
from .markers import *
from .numeric import *
from .string import *
from .window import *



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from typing import NoReturn


def reraise(
    e: Exception,
    prefix: str | None = None,
    suffix: str | None = None,
) -> NoReturn:
    class ReraisedException(type(e)):
        def __init__(self, *args):
            Exception.__init__(self, *args)

        def __getattr__(self, item):
            return getattr(e, item)

        __repr__ = Exception.__repr__
        __str__ = Exception.__str__

    ReraisedException.__name__ = type(e).__name__
    ReraisedException.__qualname__ = type(e).__qualname__
    ReraisedException.__module__ = type(e).__module__

    suffix = "" if suffix is None else suffix
    prefix = "" if prefix is None else prefix

    if suffix != "":
        suffix = "\n" + suffix

    rre = ReraisedException(f"{prefix}{e}{suffix}")
    raise rre.with_traceback(e.__traceback__) from e.__cause__



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pydiverse.transform as pdt
from pydiverse.transform import C
from pydiverse.transform._internal.pipe.verbs import (
    mutate,
    select,
)
from tests.fixtures.backend import skip_backends
from tests.util import assert_result_equal


def test_noop(df2):
    assert_result_equal(df2, lambda t: t >> mutate(col1=t.col1, col2=t.col2, col3=t.col3))


def test_multiply(df1):
    assert_result_equal(df1, lambda t: t >> mutate(x2=t.col1 * 2))
    assert_result_equal(df1, lambda t: t >> select() >> mutate(x2=t.col1 * 2))


def test_reorder(df2):
    assert_result_equal(df2, lambda t: t >> mutate(col1=t.col2, col2=t.col1))

    assert_result_equal(
        df2,
        lambda t: t >> mutate(col1=t.col2, col2=t.col1) >> mutate(col1=t.col2, col2=C.col3, col3=C.col2),
    )


def test_literals(df1):
    assert_result_equal(df1, lambda t: t >> mutate(x=1))
    assert_result_equal(df1, lambda t: t >> mutate(x=True))
    assert_result_equal(df1, lambda t: t >> mutate(x="test"))
    assert_result_equal(df1, lambda t: t >> mutate(u=pdt.lit(None)))


def test_float_lit(df1):
    assert_result_equal(df1, lambda t: t >> mutate(x=1.1))


def test_empty(df1):
    assert_result_equal(df1, lambda t: t >> mutate())


def test_none(df4):
    assert_result_equal(df4, lambda t: t >> mutate(x=None))
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x1=t.col1.is_null(),
            y1=~t.col2.is_null(),
        ),
    )


# DB2 container screws up expression reuse and fast path and-evaluation
@skip_backends("ibm_db2")
def test_mutate_bool_expr(df4):
    assert_result_equal(
        df4,
        lambda t: t >> mutate(x=t.col1 <= t.col2, y=(t.col3 * 4) >= C.col4) >> mutate(xAndY=C.x & C.y),
    )


def test_hidden_cols(df1):
    assert_result_equal(
        df1,
        lambda t: t >> select() >> mutate(col1=4) >> mutate(col1=C.col1 + t.col1),
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause


def test_version():
    import pydiverse.transform as pdt

    print(f"pdt.__version__ = \n{pdt.__version__}\n")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform.extended import *
from tests.fixtures.backend import skip_backends
from tests.util.assertion import assert_result_equal


@skip_backends("mssql", "sqlite", "ibm_db2")
def test_list_agg(df3):
    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col3) >> summarize(s=t.col2.list.agg()) >> arrange(C.s),
        check_row_order=True,
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col2)
        >> summarize(
            y=t.col5.list.agg(arrange=t.col4),
            z=t.col4.list.agg(filter=t.col1 > 0, arrange=[t.col4.descending()]),
        )
        >> arrange(t.col2),
        check_row_order=True,
    )


@skip_backends("mssql", "sqlite", "ibm_db2")
def test_list_agg_no_grouping(df3):
    assert_result_equal(
        df3,
        lambda t: t >> summarize(h=t.col5.list.agg(arrange=[t.col1, t.col4.descending()])),
        check_row_order=True,
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy
import functools
from typing import Any, Literal
from uuid import UUID

import sqlalchemy as sqa
from sqlalchemy.dialects.mssql import BIT, DATETIME2, TINYINT

from pydiverse.common import Bool, Decimal, Float, Int, String, UInt8
from pydiverse.transform._internal.backend import sql
from pydiverse.transform._internal.backend.sql import SqlImpl
from pydiverse.transform._internal.backend.targets import Target
from pydiverse.transform._internal.errors import NotSupportedError
from pydiverse.transform._internal.ops import ops
from pydiverse.transform._internal.pipe.table import Cache
from pydiverse.transform._internal.tree import types, verbs
from pydiverse.transform._internal.tree.ast import AstNode
from pydiverse.transform._internal.tree.col_expr import (
    CaseExpr,
    Cast,
    Col,
    ColExpr,
    ColFn,
    LiteralCol,
    Order,
)


class MsSqlImpl(SqlImpl):
    backend_name = "mssql"

    @classmethod
    def inf(cls):
        raise NotSupportedError("SQL Server does not support `inf`")

    @classmethod
    def nan(cls):
        raise NotSupportedError("SQL Server does not support `nan`")

    @classmethod
    def default_collation(cls):
        return "Latin1_General_bin"

    @classmethod
    def export(
        cls,
        nd: AstNode,
        target: Target,
        *,
        schema_overrides: dict[UUID, Any],
    ) -> Any:
        final_select = Cache.from_ast(nd).selected_cols()

        for col in final_select:
            if types.without_const(col.dtype()) == Bool():
                if col._uuid not in schema_overrides:
                    schema_overrides[col._uuid] = col.dtype().to_polars()

        # TODO: make sure the dtypes are preserved as in the database on export to
        # polars. Currently all ints become int64 and all floats float64
        return super().export(nd, target, schema_overrides=schema_overrides)

    @classmethod
    def build_select(cls, nd: AstNode, *, final_select: list[Col] | None = None) -> Any:
        if final_select is None:
            final_select = Cache.from_ast(nd).selected_cols()

        # boolean / bit conversion
        for desc in nd.iter_subtree_postorder():
            if isinstance(desc, verbs.Verb):
                desc.map_col_roots(
                    functools.partial(
                        convert_bool_bit,
                        desired_return_type="bool" if isinstance(desc, verbs.Filter | verbs.Join) else "bit",
                    )
                )

        # workaround for correct nulls_first / nulls_last behaviour on MSSQL
        for desc in nd.iter_subtree_postorder():
            if isinstance(desc, verbs.Arrange):
                desc.order_by = convert_order_list(desc.order_by)
            if isinstance(desc, verbs.Verb):
                for node in desc.iter_col_nodes():
                    if isinstance(node, ColFn) and (arrange := node.context_kwargs.get("arrange")):
                        node.context_kwargs["arrange"] = convert_order_list(arrange)

        sql.create_aliases(nd, {})
        table, query, sqa_expr = cls.compile_ast(nd, {col._uuid: 1 for col in final_select})

        # mssql complains about OFFSET if there is no ORDER BY
        if query.offset and not query.order_by:
            query.order_by = [Order(final_select[0])]

        return cls.compile_query(table, query, sqa_expr)

    @classmethod
    def compile_ordered_aggregation(cls, *args: sqa.ColumnElement, order_by: list[sqa.UnaryExpression], impl):
        return impl(*args).within_group(*order_by)

    @classmethod
    def fix_fn_types(cls, fn: ColFn, val: sqa.ColumnElement, *args: sqa.ColumnElement) -> sqa.ColumnElement:
        if fn.op in (ops.any, ops.all, ops.min, ops.max) and types.without_const(fn.dtype()) == Bool():
            return val.cast(BIT)

        return val

    @classmethod
    def sqa_type(cls, pdt_type):
        if isinstance(pdt_type, UInt8):
            return TINYINT
        if isinstance(pdt_type, Decimal):
            return sqa.DECIMAL(15, 6)
        return super().sqa_type(pdt_type)


def convert_order_list(order_list: list[Order]) -> list[Order]:
    new_list: list[Order] = []
    for ord in order_list:
        # is True / is False are important here since we don't want to do this costly
        # workaround if nulls_last is None (i.e. the user doesn't care)
        if ord.nulls_last is not None and (ord.nulls_last ^ ord.descending):
            new_list.append(
                Order.from_col_expr(
                    CaseExpr(
                        [(ord.order_by.is_null(), LiteralCol(int(ord.nulls_last)))],
                        LiteralCol(int(not ord.nulls_last)),
                    )
                )
            )

        new_list.append(Order(ord.order_by, ord.descending, None))

    return new_list


# MSSQL doesn't have a boolean type. This means that expressions that return a boolean
# (e.g. ==, !=, >) can't be used in other expressions without casting to the BIT type.
# Conversely, after casting to BIT, we sometimes may need to convert back to booleans.
# For any / all, we need to use max / min, which can't be applied to BIT either. Thus we
# convert to int in this case.


def convert_bool_bit(
    expr: ColExpr | Order,
    desired_return_type: Literal["bool", "bit"],
) -> ColExpr | Order:
    if isinstance(expr, Order):
        return Order(
            convert_bool_bit(expr.order_by, "bit"),
            expr.descending,
            expr.nulls_last,
        )

    result = expr

    if isinstance(expr, Col):
        return_type = "bit"

    elif isinstance(expr, ColFn):
        desired_arg_type = (
            "bool"
            if expr.op
            in (
                ops.bool_and,
                ops.bool_or,
                ops.bool_invert,
                ops.horizontal_all,
                ops.horizontal_any,
            )
            else "bit"
        )

        result = copy.copy(expr)
        result.args = list(convert_bool_bit(arg, desired_arg_type) for arg in expr.args)
        result.context_kwargs = {
            key: [convert_bool_bit(val, "bit") for val in arr] for key, arr in expr.context_kwargs.items()
        }

        return_type = (
            "bool"
            if expr.op
            in (
                ops.bool_and,
                ops.bool_or,
                ops.bool_xor,
                ops.bool_invert,
                ops.equal,
                ops.not_equal,
                ops.greater_equal,
                ops.greater_than,
                ops.less_equal,
                ops.less_than,
                ops.is_null,
                ops.is_not_null,
                ops.is_in,
                ops.str_starts_with,
                ops.str_ends_with,
                ops.str_contains,
                ops.horizontal_any,
                ops.horizontal_all,
            )
            else "bit"
        )

    elif isinstance(expr, CaseExpr):
        return_type = "bit"
        result = copy.copy(expr)
        result.cases = [(convert_bool_bit(cond, "bool"), convert_bool_bit(val, "bit")) for cond, val in expr.cases]
        result.default_val = None if expr.default_val is None else convert_bool_bit(expr.default_val, "bit")

    elif isinstance(expr, LiteralCol):
        return_type = "bit"

    elif isinstance(expr, Cast):
        return Cast(convert_bool_bit(expr.val, "bit"), expr.target_type, strict=expr.strict)

    if types.without_const(expr.dtype()) == Bool():
        if desired_return_type == "bool" and return_type == "bit":
            return ColFn(ops.equal, result, LiteralCol(True))

        if return_type == "bool" and desired_return_type == "bit":
            return CaseExpr(
                [(result, LiteralCol(True)), (~result, LiteralCol(False))],
                None,
            )

    return result


with MsSqlImpl.impl_store.impl_manager as impl:

    @impl(ops.equal, String(), String())
    def _eq(x, y):
        return (sqa.func.LENGTH(x + "a") == sqa.func.LENGTH(y + "a")) & (x.collate("Latin1_General_bin") == y)

    @impl(ops.not_equal, String(), String())
    def _ne(x, y):
        return (sqa.func.LENGTH(x + "a") != sqa.func.LENGTH(y + "a")) | (x.collate("Latin1_General_bin") != y)

    @impl(ops.less_than, String(), String())
    def _lt(x, y):
        y_ = sqa.func.SUBSTRING(y, 1, sqa.func.LENGTH(x + "a") - 1)
        return (x.collate("Latin1_General_bin") < y_) | (
            (sqa.func.LENGTH(x + "a") < sqa.func.LENGTH(y + "a")) & (x.collate("Latin1_General_bin") == y_)
        )

    @impl(ops.less_equal, String(), String())
    def _le(x, y):
        y_ = sqa.func.SUBSTRING(y, 1, sqa.func.LENGTH(x + "a") - 1)
        return (x.collate("Latin1_General_bin") < y_) | (
            (sqa.func.LENGTH(x + "a") <= sqa.func.LENGTH(y + "a")) & (x.collate("Latin1_General_bin") == y_)
        )

    @impl(ops.greater_than, String(), String())
    def _gt(x, y):
        y_ = sqa.func.SUBSTRING(y, 1, sqa.func.LENGTH(x + "a") - 1)
        return (x.collate("Latin1_General_bin") > y_) | (
            (sqa.func.LENGTH(x + "a") > sqa.func.LENGTH(y + "a")) & (x.collate("Latin1_General_bin") == y_)
        )

    @impl(ops.greater_equal, String(), String())
    def _ge(x, y):
        y_ = sqa.func.SUBSTRING(y, 1, sqa.func.LENGTH(x + "a") - 1)
        return (x.collate("Latin1_General_bin") > y_) | (
            (sqa.func.LENGTH(x + "a") >= sqa.func.LENGTH(y + "a")) & (x.collate("Latin1_General_bin") == y_)
        )

    @impl(ops.str_len)
    def _str_length(x):
        return sqa.func.LENGTH(x + "a", type_=sqa.BigInteger()) - 1

    @impl(ops.str_replace_all)
    def _str_replace_all(x, y, z):
        x = x.collate(MsSqlImpl.default_collation())
        return sqa.func.REPLACE(x, y, z, type_=x.type)

    @impl(ops.str_starts_with)
    def _str_starts_with(x, y):
        x = x.collate(MsSqlImpl.default_collation())
        return x.startswith(y, autoescape=True)

    @impl(ops.str_ends_with)
    def _str_ends_with(x, y):
        x = x.collate(MsSqlImpl.default_collation())
        return x.endswith(y, autoescape=True)

    @impl(ops.str_contains)
    def _str_contains(x, pattern, allow_regex, true_if_regex_unsupported):
        if true_if_regex_unsupported:
            return sqa.literal(True, literal_execute=True)
        x = x.collate(MsSqlImpl.default_collation())
        return x.contains(pattern, autoescape=True)

    @impl(ops.str_slice)
    def _str_slice(x, offset, length):
        return sqa.func.SUBSTRING(x, offset + 1, length)

    @impl(ops.dt_day_of_week)
    def _dt_day_of_week(x):
        # Offset DOW such that Mon=1, Sun=7
        _1 = sqa.literal_column("1")
        _2 = sqa.literal_column("2")
        _7 = sqa.literal_column("7")
        return (sqa.extract("dow", x) + sqa.text("@@DATEFIRST") - _2) % _7 + _1

    @impl(ops.mean)
    def _mean(x):
        return sqa.func.AVG(sqa.cast(x, sqa.Double()), type_=sqa.Double())

    @impl(ops.log)
    def _log(x):
        return sqa.func.log(x)

    @impl(ops.ceil)
    def _ceil(x):
        return sqa.func.ceiling(x)

    @impl(ops.str_to_datetime)
    def _str_to_datetime(x):
        return sqa.cast(x, DATETIME2)

    @impl(ops.is_inf)
    def _is_inf(x):
        return False

    @impl(ops.is_not_inf)
    def _is_not_inf(x):
        return True

    @impl(ops.is_nan)
    def _is_nan(x):
        return False

    @impl(ops.is_not_nan)
    def _is_not_nan(x):
        return True

    @impl(ops.pow)
    def _pow(x, y):
        return_type = sqa.Double()
        if isinstance(x.type, sqa.Numeric) and isinstance(y.type, sqa.Numeric):
            return_type = sqa.Numeric()
        return sqa.func.POWER(x, y, type_=return_type)

    @impl(ops.pow, Int(), Int())
    def _pow_int(x, y):
        return sqa.func.POWER(sqa.cast(x, type_=sqa.Double()), y)

    @impl(ops.all)
    @impl(ops.min, Bool())
    def _all(x):
        return sqa.func.min(sqa.cast(x, TINYINT))

    @impl(ops.any)
    @impl(ops.max, Bool())
    def _all(x):
        return sqa.func.max(sqa.cast(x, TINYINT))

    @impl(ops.truediv, Int(), Int())
    @impl(ops.truediv, Float(), Float())
    def _truediv(x, y):
        if not isinstance(x.type, Float):
            x = sqa.cast(x, sqa.Double)
        if not isinstance(y.type, Float):
            y = sqa.cast(y, sqa.Double)
        return x / y

    @impl(ops.cbrt)
    def _cbrt(x):
        return sqa.func.sign(x) * _pow(sqa.func.abs(x), sqa.literal(1 / 3, type_=sqa.Double))

    @impl(ops.rand)
    def _rand():
        return sqa.func.RAND(sqa.text("convert(varbinary, newid())"))



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

from pydiverse.common import Int
from pydiverse.transform._internal.ops.op import ContextKwarg, Ftype, Operator
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import NUMERIC, Const, S


class Window(Operator):
    def __init__(
        self,
        name: str,
        *signatures: Signature,
        param_names: list[str] | None = None,
        default_values: list[Any] | None = None,
        generate_expr_method=False,
        arrange_required=True,
        doc: str = "",
    ):
        super().__init__(
            name,
            *signatures,
            ftype=Ftype.WINDOW,
            context_kwargs=[
                ContextKwarg("partition_by", False),
                ContextKwarg("arrange", arrange_required),
            ],
            param_names=param_names,
            default_values=default_values,
            generate_expr_method=generate_expr_method,
            doc=doc,
        )


shift = Window(
    "shift",
    Signature(S, Const(Int()), Const(S), return_type=S),
    param_names=["self", "n", "fill_value"],
    default_values=[..., ..., None],
    generate_expr_method=True,
    arrange_required=False,
    doc="""
Shifts values in the column by an offset.

:param n:
    The number of places to shift by. May be negative.

:param fill_value:
    The value to write to the empty spaces created by the shift. Defaults to
    null.

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": [5, -1, 435, -34, 8, None, 0],
...         "b": ["r", "True", "??", ".  .", "-1/12", "abc", "#"],
...     }
... )
>>> (
...     t
...     >> mutate(
...         x=t.a.shift(2, -40),
...         y=t.b.shift(1, arrange=t.a.nulls_last()),
...     )
...     >> show()
... )
Table <unnamed>, backend: PolarsImpl
shape: (7, 4)
┌──────┬───────┬─────┬───────┐
│ a    ┆ b     ┆ x   ┆ y     │
│ ---  ┆ ---   ┆ --- ┆ ---   │
│ i64  ┆ str   ┆ i64 ┆ str   │
╞══════╪═══════╪═════╪═══════╡
│ 5    ┆ r     ┆ -40 ┆ #     │
│ -1   ┆ True  ┆ -40 ┆ .  .  │
│ 435  ┆ ??    ┆ 5   ┆ -1/12 │
│ -34  ┆ .  .  ┆ -1  ┆ null  │
│ 8    ┆ -1/12 ┆ 435 ┆ r     │
│ null ┆ abc   ┆ -34 ┆ ??    │
│ 0    ┆ #     ┆ 8   ┆ True  │
└──────┴───────┴─────┴───────┘
""",
)

row_number = Window(
    "row_number",
    Signature(return_type=Int()),
    arrange_required=False,
    doc="""
Computes the index of a row.

Via the *arrange* argument, this can be done relative to a different order of
the rows. But note that the result may not be unique if the argument of
*arrange* contains duplicates.

Examples
--------
>>> t = pdt.Table({"a": [5, -1, 435, -34, 8, None, 0]})
>>> (
...     t
...     >> mutate(
...         x=pdt.row_number(),
...         y=pdt.row_number(arrange=t.a),
...     )
...     >> show()
... )
Table <unnamed>, backend: PolarsImpl
shape: (7, 3)
┌──────┬─────┬─────┐
│ a    ┆ x   ┆ y   │
│ ---  ┆ --- ┆ --- │
│ i64  ┆ i64 ┆ i64 │
╞══════╪═════╪═════╡
│ 5    ┆ 1   ┆ 5   │
│ -1   ┆ 2   ┆ 3   │
│ 435  ┆ 3   ┆ 7   │
│ -34  ┆ 4   ┆ 2   │
│ 8    ┆ 5   ┆ 6   │
│ null ┆ 6   ┆ 1   │
│ 0    ┆ 7   ┆ 4   │
└──────┴─────┴─────┘
""",
)

rank = Window(
    "rank",
    Signature(return_type=Int()),
    doc="""
The number of strictly smaller elements in the column plus one.

This is the same as ``rank("min")`` in polars. This function has two syntax
alternatives, as shown in the example below. The pdt. version is a bit more
flexible, because it allows sorting by multiple expressions.


Examples
--------
>>> t = pdt.Table({"a": [5, -1, 435, -1, 8, None, 8]})
>>> (
...     t
...     >> mutate(
...         x=t.a.nulls_first().rank(),
...         y=pdt.rank(arrange=t.a.nulls_first()),
...     )
...     >> show()
... )
Table <unnamed>, backend: PolarsImpl
shape: (7, 3)
┌──────┬─────┬─────┐
│ a    ┆ x   ┆ y   │
│ ---  ┆ --- ┆ --- │
│ i64  ┆ i64 ┆ i64 │
╞══════╪═════╪═════╡
│ 5    ┆ 4   ┆ 4   │
│ -1   ┆ 2   ┆ 2   │
│ 435  ┆ 7   ┆ 7   │
│ -1   ┆ 2   ┆ 2   │
│ 8    ┆ 5   ┆ 5   │
│ null ┆ 1   ┆ 1   │
│ 8    ┆ 5   ┆ 5   │
└──────┴─────┴─────┘
""",
)

dense_rank = Window(
    "dense_rank",
    Signature(return_type=Int()),
    doc="""
The number of smaller or equal values in the column (not counting duplicates).

This function has two syntax alternatives, as shown in the example below. The
pdt. version is a bit more flexible, because it allows sorting by multiple
expressions.

Examples
--------
>>> t = pdt.Table({"a": [5, -1, 435, -1, 8, None, 8]})
>>> (
...     t
...     >> mutate(
...         x=t.a.nulls_first().dense_rank(),
...         y=pdt.dense_rank(arrange=t.a.nulls_first()),
...     )
...     >> show()
... )
Table <unnamed>, backend: PolarsImpl
shape: (7, 3)
┌──────┬─────┬─────┐
│ a    ┆ x   ┆ y   │
│ ---  ┆ --- ┆ --- │
│ i64  ┆ i64 ┆ i64 │
╞══════╪═════╪═════╡
│ 5    ┆ 3   ┆ 3   │
│ -1   ┆ 2   ┆ 2   │
│ 435  ┆ 5   ┆ 5   │
│ -1   ┆ 2   ┆ 2   │
│ 8    ┆ 4   ┆ 4   │
│ null ┆ 1   ┆ 1   │
│ 8    ┆ 4   ┆ 4   │
└──────┴─────┴─────┘
""",
)


cum_sum = Window(
    "cum_sum",
    *(Signature(dtype, return_type=dtype) for dtype in NUMERIC),
    generate_expr_method=True,
    doc="""
The sum of all preceding elements and the current element.

Null values are assigned the sum of all preceding elements.
""",
)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import functools
import operator
import uuid
from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Any
from uuid import UUID

import polars as pl
import sqlalchemy as sqa

from pydiverse.transform._internal.backend.impl_store import ImplStore
from pydiverse.transform._internal.backend.targets import Target
from pydiverse.transform._internal.errors import NotSupportedError
from pydiverse.transform._internal.ops import ops
from pydiverse.transform._internal.ops.op import Ftype
from pydiverse.transform._internal.tree.ast import AstNode
from pydiverse.transform._internal.tree.col_expr import Col, ColFn, LiteralCol
from pydiverse.transform._internal.tree.types import Dtype

try:
    import pandas as pd
except ImportError:
    pd = None

if TYPE_CHECKING:
    from pydiverse.transform._internal.ops.ops import Operator


class TableImpl(AstNode):
    backend_name: str
    impl_store = ImplStore()

    def __init__(self, name: str | None, schema: dict[str, Dtype]):
        self.name = name
        self.cols = {name: Col(name, self, uuid.uuid1(), dtype, Ftype.ELEMENT_WISE) for name, dtype in schema.items()}

    def _unformatted_ast_repr(self, verb_depth: int, expr_depth: int, display_name_map) -> str:
        return self._ast_node_repr(expr_depth, display_name_map)

    def _ast_node_repr(self, expr_depth, display_name_map):
        return display_name_map[self]

    def _table_def_repr(self) -> str:
        raise NotImplementedError()

    def short_name(self):
        return ("?" if self.name is None else self.name) + f" (source table, backend: '{self.backend_name}')"

    def __init_subclass__(cls) -> None:
        cls.impl_store = ImplStore()

    @staticmethod
    def from_resource(
        resource: Any,
        backend: Target | None = None,
        *,
        name: str | None = None,
        uuids: dict[str, UUID] | None = None,
    ) -> "TableImpl":
        from pydiverse.transform._internal.backend.targets import (
            DuckDb,
            Polars,
            SqlAlchemy,
        )

        if isinstance(resource, AstNode):
            res = resource

        elif isinstance(resource, dict):
            return TableImpl.from_resource(pl.DataFrame(resource), backend, name=name, uuids=uuids)

        elif pd is not None and isinstance(resource, pd.DataFrame):
            # copy pandas dataframe to polars
            # TODO: try zero-copy for arrow backed pandas
            return TableImpl.from_resource(pl.DataFrame(resource), backend, name=name, uuids=uuids)

        elif isinstance(resource, pl.DataFrame | pl.LazyFrame):
            if name is None:
                # If the data frame has be previously exported by transform, a
                # name attribute was added.
                if hasattr(resource, "name"):
                    name = resource.name
            if backend is None or isinstance(backend, Polars):
                from pydiverse.transform._internal.backend.polars import PolarsImpl

                res = PolarsImpl(name, resource)
            elif isinstance(backend, DuckDb):
                from pydiverse.transform._internal.backend.duckdb_polars import (
                    DuckDbPolarsImpl,
                )

                res = DuckDbPolarsImpl(name, resource)

        elif isinstance(resource, str | sqa.Table):
            if not isinstance(backend, SqlAlchemy):
                raise TypeError(
                    "If `resource` is a string or a SQLAlchemy table, `backend` must "
                    "have type `SqlALchemy` and contain an engine."
                )

            from pydiverse.transform._internal.backend.sql import SqlImpl

            res = SqlImpl(resource, backend, name)

        else:
            raise AssertionError

        if uuids is not None:
            for name, col in res.cols.items():
                col._uuid = uuids[name]

        return res

    def iter_subtree_postorder(self) -> Iterable[AstNode]:
        yield self

    def iter_subtree_preorder(self) -> Iterable[AstNode]:
        yield self

    @classmethod
    def build_query(cls, nd: AstNode) -> str | None: ...

    @classmethod
    def export(
        cls,
        nd: AstNode,
        target: Target,
        *,
        schema_overrides: dict[UUID, Any],
    ) -> Any: ...

    @classmethod
    def get_impl(cls, op: "Operator", sig: Sequence[Dtype]) -> Any:
        if (impl := cls.impl_store.get_impl(op, sig)) is not None:
            return impl

        if cls is TableImpl:
            raise NotSupportedError

        try:
            return cls.__bases__[0].get_impl(op, sig)
        except NotSupportedError as err:
            raise NotSupportedError(
                f"operation `{op.name}` is not supported by the backend `{cls.__name__.lower()[:-4]}`"
            ) from err


def split_join_cond(on: ColFn) -> list[ColFn]:
    if isinstance(on, LiteralCol):
        return []
    elif on.op == ops.bool_and:
        return split_join_cond(on.args[0]) + split_join_cond(on.args[1])
    elif on.op == ops.horizontal_all:
        return functools.reduce(operator.add, (split_join_cond(arg) for arg in on.args))
    else:
        return [on]


# Returns the left and right columns of a list of equality predicates.
def get_left_right_on(
    eq_predicates: list[ColFn], left_uuids: set[UUID], right_uuids: set[UUID]
) -> tuple[list[Col], list[Col]] | None:
    left_on = []
    right_on = []
    for pred in eq_predicates:
        left_on.append(pred.args[0])
        right_on.append(pred.args[1])

        must_swap_cols = None
        for e in pred.args[0].iter_subtree_postorder():
            if isinstance(e, Col):
                must_swap_cols = e._uuid in right_uuids
                assert must_swap_cols or e._uuid in left_uuids
                break

        assert must_swap_cols is not None

        if must_swap_cols:
            left_on[-1], right_on[-1] = right_on[-1], left_on[-1]

    return (left_on, right_on)


with TableImpl.impl_store.impl_manager as impl:

    @impl(ops.add)
    def _add(lhs, rhs):
        return lhs + rhs

    @impl(ops.sub)
    def _sub(lhs, rhs):
        return lhs - rhs

    @impl(ops.mul)
    def _mul(lhs, rhs):
        return lhs * rhs

    @impl(ops.truediv)
    def _truediv(lhs, rhs):
        return lhs / rhs

    @impl(ops.floordiv)
    def _floordiv(lhs, rhs):
        return lhs // rhs

    @impl(ops.pow)
    def _pow(lhs, rhs):
        return lhs**rhs

    @impl(ops.mod)
    def _mod(lhs, rhs):
        return lhs % rhs

    @impl(ops.neg)
    def _neg(x):
        return -x

    @impl(ops.pos)
    def _pos(x):
        return +x

    @impl(ops.abs)
    def _abs(x):
        return abs(x)

    @impl(ops.bool_and)
    def _and(lhs, rhs):
        return lhs & rhs

    @impl(ops.bool_or)
    def _or(lhs, rhs):
        return lhs | rhs

    @impl(ops.bool_xor)
    def _xor(lhs, rhs):
        return lhs ^ rhs

    @impl(ops.bool_invert)
    def _invert(x):
        return ~x

    @impl(ops.equal)
    def _eq(lhs, rhs):
        return lhs == rhs

    @impl(ops.not_equal)
    def _ne(lhs, rhs):
        return lhs != rhs

    @impl(ops.less_than)
    def _lt(lhs, rhs):
        return lhs < rhs

    @impl(ops.less_equal)
    def _le(lhs, rhs):
        return lhs <= rhs

    @impl(ops.greater_than)
    def _gt(lhs, rhs):
        return lhs > rhs

    @impl(ops.greater_equal)
    def _ge(lhs, rhs):
        return lhs >= rhs

    @impl(ops.horizontal_all)
    def _horizontal_all(*args):
        return functools.reduce(_and, args)

    @impl(ops.horizontal_any)
    def _horizontal_any(*args):
        return functools.reduce(_or, args)

    @impl(ops.horizontal_sum)
    def _horizontal_sum(*args):
        return functools.reduce(_add, args)
